"""A statistics record for the simulation."""

from itertools import chain
from math import isfinite, isnan, nan
from typing import Final, Generator, Iterable

from moptipy.utils.logger import KEY_VALUE_SEPARATOR
from pycommons.io.csv import CSV_SEPARATOR
from pycommons.math.stream_statistics import StreamStats
from pycommons.strings.string_conv import num_to_str
from pycommons.types import check_int_range

#: the name of the statistics key
COL_STAT: Final[str] = "stat"
#: the total column name
COL_TOTAL: Final[str] = "total"
#: the product column prefix
COL_PRODUCT_PREFIX: Final[str] = "product_"
#: the mean TRP row
ROW_TRP_MEAN: Final[str] = "trp.mean"
#: the TRP standard deviation row
ROW_TRP_SD: Final[str] = "trp.sd"
#: the fill rate row
ROW_FILL_RATE: Final[str] = "fill.rate"
#: the mean CWT row
ROW_CWT_MEAN: Final[str] = "cwt.mean"
#: the CWT standard deviation row
ROW_CWT_SD: Final[str] = "cwt.sd"
#: the mean stock level row
ROW_STOCK_LEVEL_MEAN: Final[str] = "stocklevel.mean"
#: the fulfilled rate
ROW_FULFILLED_RATE: Final[str] = "fulfilled.rate"
ROW_SIMULATION_TIME: Final[str] = f"time.s{KEY_VALUE_SEPARATOR}"


class Statistics:
    """A statistics record based on production scheduling."""

    def __init__(self, n_products: int) -> None:
        """
        Create the statistics record.

        :param n_products: the number of products
        """
        check_int_range(n_products, "n_products", 1, 1_000_000_000)
        #: the production time means per-product
        self.production_time_means: Final[list[float]] = [nan] * n_products
        #: the overall production time mean
        self.production_time_mean: float = nan
        #: the production time standard deviations per-product
        self.production_time_sds: Final[list[float]] = [nan] * n_products
        #: the overall production time standard deviations
        self.production_time_sd: float = nan
        #: the fraction of demands that were immediately satisfied,
        #: on a per-product basis
        self.immediately_satisfied_rates: Final[list[float]] = (
            [nan] * n_products)
        #: the overall fraction of immediately satisfied demands
        self.immediately_satisfied_rate: float = nan
        #: the average waiting time for all demands that were not immediately
        #: satisfied -- only counting demands that were actually satisfied
        self.waiting_time_for_none_immediates_means: Final[list[float]] = (
            [nan] * n_products)
        #: the standard deviation of the waiting time for all demands that
        #: were not immediately satisfied -- only counting demands that were
        #: actually satisfied
        self.waiting_time_for_none_immediates_sds: Final[list[float]] = (
            [nan] * n_products)
        #: the overall average waiting time for all demands that were not
        #: immediately satisfied -- only counting demands that were actually
        #: satisfied
        self.waiting_time_for_none_immediates_mean: float = nan
        #: the overall standard deviation of the waiting time for all demands
        #: that were not immediately satisfied -- only counting demands that
        #: were actually  satisfied
        self.waiting_time_for_none_immediates_sd: float = nan
        #: the fraction of demands that were fulfilled, on a per-product basis
        self.fulfilled_rates: Final[list[float]] = [nan] * n_products
        #: the fraction of demands that were fulfilled overall
        self.fulfilled_rate: float = nan
        #: the average stock level, on a per-product basis
        self.stock_level_means: Final[list[float]] = [nan] * n_products
        #: the overall average stock level
        self.stock_level_mean: float = nan
        #: the nano seconds used by the simulation
        self.simulation_time_nanos: int | float = nan

    def __str__(self) -> str:
        """Convert this object to a string."""
        return "\n".join(to_stream(self))


def __nts(num: int | float) -> str:
    """
    Convert a number to a string, taking care of `nan`.

    :param num: the number
    :return: the string
    """
    return "nan" if isnan(num) else num_to_str(num)


def to_stream(stats: Statistics) -> Generator[str, None, None]:
    """
    Write a statistics record to a stream.

    :param stats: the statistics record
    :return: the stream of data
    """
    n_products: Final[int] = list.__len__(stats.production_time_means)
    yield str.join(CSV_SEPARATOR, chain((
        COL_STAT, COL_TOTAL), (f"{COL_PRODUCT_PREFIX}{i}" for i in range(
            n_products))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_TRP_MEAN, __nts(stats.production_time_mean)), (
        map(__nts, stats.production_time_means))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_TRP_SD, __nts(stats.production_time_sd)), (
        map(__nts, stats.production_time_sds))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_FILL_RATE, __nts(stats.immediately_satisfied_rate)), (
        map(__nts, stats.immediately_satisfied_rates))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_CWT_MEAN, __nts(stats.waiting_time_for_none_immediates_mean)), (
        map(__nts, stats.waiting_time_for_none_immediates_means))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_CWT_SD, __nts(stats.waiting_time_for_none_immediates_sd)), (
        map(__nts, stats.waiting_time_for_none_immediates_sds))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_STOCK_LEVEL_MEAN, __nts(stats.stock_level_mean)), (
        map(__nts, stats.stock_level_means))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_FULFILLED_RATE, __nts(stats.fulfilled_rate)), (
        map(__nts, stats.fulfilled_rates))))
    yield f"{ROW_SIMULATION_TIME}{__nts(
        stats.simulation_time_nanos / 1_000_000_000)}"


def mean(stats: Iterable[Statistics]) -> Statistics:
    """
    Compute the mean over a given statistic.

    :param stats: the statistic
    :return: the mean statistics
    """
    stats = tuple(stats)
    n_products = list.__len__(stats[0].production_time_means)
    for stat in stats:
        if list.__len__(stat.production_time_means) != n_products:
            raise ValueError("Inconsistent number of products.")

    result: Final[Statistics] = Statistics(n_products)
    ss: Final[StreamStats] = StreamStats()

    for i in range(n_products):
        ss.reset()
        ss.update(stat.stock_level_means[i] for stat in stats if isfinite(
            stat.stock_level_means[i]))
        result.stock_level_means[i] = ss.mean()
    ss.reset()
    ss.update(stat.stock_level_mean for stat in stats if isfinite(
        stat.stock_level_mean))
    result.stock_level_mean = ss.mean()

    for i in range(n_products):
        ss.reset()
        ss.update(stat.production_time_means[i] for stat in stats if isfinite(
            stat.production_time_means[i]))
        result.production_time_means[i] = ss.mean()
    ss.reset()
    ss.update(stat.production_time_mean for stat in stats if isfinite(
        stat.production_time_mean))
    result.production_time_mean = ss.mean()

    for i in range(n_products):
        ss.reset()
        ss.update(stat.production_time_sds[i] for stat in stats if isfinite(
            stat.production_time_sds[i]))
        result.production_time_sds[i] = ss.mean()
    ss.reset()
    ss.update(stat.production_time_sd for stat in stats if isfinite(
        stat.production_time_sd))
    result.production_time_sd = ss.mean()

    for i in range(n_products):
        ss.reset()
        ss.update(stat.immediately_satisfied_rates[i]
                  for stat in stats if isfinite(
            stat.immediately_satisfied_rates[i]))
        result.immediately_satisfied_rates[i] = ss.mean()
    ss.reset()
    ss.update(stat.immediately_satisfied_rate for stat in stats if isfinite(
        stat.immediately_satisfied_rate))
    result.immediately_satisfied_rate = ss.mean()

    for i in range(n_products):
        ss.reset()
        ss.update(stat.waiting_time_for_none_immediates_means[i]
                  for stat in stats if isfinite(
            stat.waiting_time_for_none_immediates_means[i]))
        result.waiting_time_for_none_immediates_means[i] = ss.mean()
    ss.reset()
    ss.update(stat.waiting_time_for_none_immediates_mean
              for stat in stats if isfinite(
                  stat.waiting_time_for_none_immediates_mean))
    result.waiting_time_for_none_immediates_mean = ss.mean()

    for i in range(n_products):
        ss.reset()
        ss.update(stat.waiting_time_for_none_immediates_sds[i]
                  for stat in stats if isfinite(
            stat.waiting_time_for_none_immediates_sds[i]))
        result.waiting_time_for_none_immediates_sds[i] = ss.mean()
    ss.reset()
    ss.update(stat.waiting_time_for_none_immediates_sd
              for stat in stats if isfinite(
                  stat.waiting_time_for_none_immediates_sd))
    result.waiting_time_for_none_immediates_sd = ss.mean()

    for i in range(n_products):
        ss.reset()
        ss.update(stat.fulfilled_rates[i] for stat in stats if isfinite(
            stat.fulfilled_rates[i]))
        result.fulfilled_rates[i] = ss.mean()
    ss.reset()
    ss.update(stat.fulfilled_rate for stat in stats if isfinite(
        stat.fulfilled_rate))
    result.fulfilled_rate = ss.mean()

    # get the mean simulation time: most likely a sum of integers
    count: int = 0
    overall: int | float = 0
    for stat in stats:
        val: int | float = stat.simulation_time_nanos
        if isfinite(val):
            count += 1
            overall += val
    result.simulation_time_nanos = nan if count <= 0 else overall / count

    return result



"""
A production scheduling instance.

Production instances have names :attr:`Instance.name`.

Notice that production times are used in a cycling fashion.
The time when a certain product is finished can be computed via
:func:`~compute_finish_time` in an efficient way.

>>> name = "my_instance"

The number of products be 3.
>>> n_products = 3

The number of customers be 5.
>>> n_customers = 5

The number of stations be 4.
>>> n_stations = 4

There will be 6 customer demands.
>>> n_demands = 6

The end of the warmup period.
>>> time_end_warmup = 10

The end of the measurement period.
>>> time_end_measure = 10000

Each product may take a different route through different stations.
>>> route_p0 = [0, 3, 2]
>>> route_p1 = [0, 2, 1, 3]
>>> route_p2 = [1, 2, 3]
>>> routes = [route_p0, route_p1, route_p2]

Each demand is a tuple of demand_id, customer_id, product_id, amount,
release time, and deadline.
>>> d0 = [0, 0, 1, 20, 1240,  3000]
>>> d1 = [1, 1, 0, 10, 2300,  4000]
>>> d2 = [2, 2, 2,  7, 8300, 11000]
>>> d3 = [3, 3, 1, 12, 7300,  9000]
>>> d4 = [4, 4, 2, 23, 5410, 16720]
>>> d5 = [5, 3, 0, 19, 4234, 27080]
>>> demands = [d0, d1, d2, d3, d4, d5]

There is a fixed amount of each product in the warehouse at time step 0.
>>> warehous_at_t0 = [10, 0, 6]

Each station requires a certain working time for each unit of each product.
This production time may vary over time.
For example, maybe station 0 needs 10 time units for 1 unit of product 0 from
time step 0 to time step 19, then 11 time units from time step 20 to 39, then
8 time units from time step 40 to 59.
These times are cyclic, meaning that at time step 60 to 79, it will again need
10 time units, and so on.
Of course, production times are only specified for stations that a product is
actually routed through.
>>> m0_p0 = [10.0, 20.0, 11.0, 40.0,  8.0, 60.0]
>>> m0_p1 = [12.0, 20.0,  7.0, 40.0, 11.0, 70.0]
>>> m0_p2 = []
>>> m1_p0 = []
>>> m1_p1 = [20.0, 50.0, 30.0, 120.0,  7.0, 200.0]
>>> m1_p2 = [21.0, 50.0, 29.0, 130.0,  8.0, 190.0]
>>> m2_p0 = [ 8.0, 20.0,  9.0, 60.0]
>>> m2_p1 = [10.0, 90.0]
>>> m2_p2 = [12.0, 70.0,  30.0, 120.0]
>>> m3_p0 = [70.0, 200.0,  3.0, 220.0]
>>> m3_p1 = [60.0, 220.0,  5.0, 260.0]
>>> m3_p2 = [30.0, 210.0, 10.0, 300.0]
>>> station_product_unit_times = [[m0_p0, m0_p1, m0_p2],
...                               [m1_p0, m1_p1, m1_p2],
...                               [m2_p0, m2_p1, m2_p2],
...                               [m3_p0, m3_p1, m3_p2]]

We can (but do not need to) provide additional information as key-value pairs.
>>> infos = {"source": "manually created",
...          "creation_date": "2025-11-09"}

From all of this data, we can create the instance.
>>> instance = Instance(name, n_products, n_customers, n_stations, n_demands,
...                     time_end_warmup, time_end_measure,
...                     routes, demands, warehous_at_t0,
...                     station_product_unit_times, infos)
>>> instance.name
'my_instance'

>>> instance.n_customers
5

>>> instance.n_stations
4

>>> instance.n_demands
6

>>> instance.n_products
3

>>> instance.routes
((0, 3, 2), (0, 2, 1, 3), (1, 2, 3))

>>> instance.time_end_warmup
10.0

>>> instance.time_end_measure
10000.0

>>> instance.demands
(Demand(arrival=1240.0, deadline=3000.0, demand_id=0, customer_id=0,\
 product_id=1, amount=20, measure=True),\
 Demand(arrival=2300.0, deadline=4000.0, demand_id=1, customer_id=1,\
 product_id=0, amount=10, measure=True),\
 Demand(arrival=4234.0, deadline=27080.0, demand_id=5, customer_id=3,\
 product_id=0, amount=19, measure=True),\
 Demand(arrival=5410.0, deadline=16720.0, demand_id=4, customer_id=4,\
 product_id=2, amount=23, measure=True),\
 Demand(arrival=7300.0, deadline=9000.0, demand_id=3, customer_id=3,\
 product_id=1, amount=12, measure=True),\
 Demand(arrival=8300.0, deadline=11000.0, demand_id=2, customer_id=2,\
 product_id=2, amount=7, measure=True))

>>> instance.warehous_at_t0
(10, 0, 6)

>>> instance.station_product_unit_times
((array([10., 20., 11., 40.,  8., 60.]), \
array([12., 20.,  7., 40., 11., 70.]), array([], dtype=float64)), (\
array([], dtype=float64), array([ 20.,  50.,  30., 120.,   7., 200.]), \
array([ 21.,  50.,  29., 130.,   8., 190.])), (array([ 8., 20.,  9., 60.]), \
array([10., 90.]), array([ 12.,  70.,  30., 120.])), (\
array([ 70., 200.,   3., 220.]), array([ 60., 220.,   5., 260.]), array(\
[ 30., 210.,  10., 300.])))

>>> instance.n_measurable_demands
6

>>> instance.n_measurable_demands_per_product
(2, 2, 2)

>>> dict(instance.infos)
{'source': 'manually created', 'creation_date': '2025-11-09'}

We can serialize instances to a stream of strings and also load them back
from a stream of strings.
Here, we store `instance` to a stream.
We then load the independent instance `i2` from that stream.
>>> i2 = from_stream(to_stream(instance))
>>> i2 is instance
False
>>> i2 == instance
True

You can see that the loaded instance has the same data as the stored one.
>>> i2.name == instance.name
True
>>> i2.n_customers == instance.n_customers
True
>>> i2.n_stations == instance.n_stations
True
>>> i2.n_demands == instance.n_demands
True
>>> i2.n_products == instance.n_products
True
>>> i2.routes == instance.routes
True
>>> i2.demands == instance.demands
True
>>> i2.time_end_warmup == instance.time_end_warmup
True
>>> i2.time_end_measure == instance.time_end_measure
True
>>> i2.warehous_at_t0 == instance.warehous_at_t0
True
>>> eq: bool = True
>>> for i in range(i2.n_stations):
...     ma1 = i2.station_product_unit_times[i]
...     ma2 = instance.station_product_unit_times[i]
...     for j in range(i2.n_products):
...         pr1 = ma1[j]
...         pr2 = ma2[j]
...         if not np.array_equal(pr1, pr2):
...             eq = False
>>> eq
True

True
>>> i2.infos == instance.infos
True
"""

from dataclasses import dataclass
from itertools import batched
from math import ceil, isfinite
from string import ascii_letters, digits
from typing import (
    Callable,
    Final,
    Generator,
    Iterable,
    Iterator,
    Mapping,
    cast,
)

import numba  # type: ignore
import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import (
    COMMENT_START,
    KEY_VALUE_SEPARATOR,
)
from moptipy.utils.strings import sanitize_name
from pycommons.ds.cache import repr_cache
from pycommons.ds.immutable_map import immutable_mapping
from pycommons.io.csv import CSV_SEPARATOR
from pycommons.math.int_math import try_int
from pycommons.strings.string_conv import bool_to_str, float_to_str, num_to_str
from pycommons.types import check_int_range, check_to_int_range, type_error

#: The maximum for the number of stations, products, or customers.
MAX_ID: Final[int] = 1_000_000_000

#: No value bigger than this is permitted in any tuple anywhere.
MAX_VALUE: Final[int] = 2_147_483_647

#: the index of the demand ID
DEMAND_ID: Final[int] = 0
#: the index of the customer ID
DEMAND_CUSTOMER: Final[int] = 1
#: the index of the product ID
DEMAND_PRODUCT: Final[int] = 2
#: the index of the demanded amount
DEMAND_AMOUNT: Final[int] = 3
#: the index of the demand release time
DEMAND_ARRIVAL: Final[int] = 4
#: the index of the demand deadline
DEMAND_DEADLINE: Final[int] = 5


@dataclass(order=True, frozen=True)
class Demand(Iterable[int | float]):
    """
    The record for demands.

    >>> Demand(arrival=0.6, deadline=0.8, demand_id=1,
    ...        customer_id=2, product_id=6, amount=12, measure=True)
    Demand(arrival=0.6, deadline=0.8, demand_id=1, customer_id=2,\
 product_id=6, amount=12, measure=True)
    >>> Demand(arrival=16, deadline=28, demand_id=1,
    ...        customer_id=2, product_id=6, amount=12, measure=False)
    Demand(arrival=16.0, deadline=28.0, demand_id=1, customer_id=2,\
 product_id=6, amount=12, measure=False)
    """

    #: the arrival time, i.e., when the demand enters the system
    arrival: float
    #: the deadline, i.e., when the customer expects the result
    deadline: float
    #: the ID of the demand
    demand_id: int
    #: the customer ID
    customer_id: int
    #: the ID of the product
    product_id: int
    #: the amount
    amount: int
    #: is this demand measurement relevant?
    measure: bool

    def __init__(self, arrival: int | float,
                 deadline: int | float, demand_id: int,
                 customer_id: int, product_id: int, amount: int,
                 measure: bool) -> None:
        """
        Initialize the record.

        :param arrival: the arrival time
        :param deadline: the deadline
        :param demand_id: the demand id
        :param customer_id: the customer id
        :param product_id: the product id
        :param amount: the amount
        :param measure: is this demand relevant for measurement?
        """
        if isinstance(arrival, int):
            t: float = float(arrival)
            if t != arrival:
                raise ValueError(f"invalid arrival time {arrival}")
            arrival = t
        if not isinstance(arrival, float):
            raise type_error(arrival, "arrival", float)
        if not (isfinite(arrival) and (
                0 < arrival < MAX_VALUE)):
            raise ValueError(f"invalid arrival={arrival}")

        if isinstance(deadline, int):
            t = float(deadline)
            if t != deadline:
                raise ValueError(f"invalid deadline time {deadline}")
            deadline = t
        if not isinstance(deadline, float):
            raise type_error(deadline, "deadline", float)
        if not (isfinite(deadline) and (0 < deadline < MAX_VALUE)):
            raise ValueError(f"invalid deadline={deadline}")

        if deadline < arrival:
            raise ValueError(
                f"arrival={arrival} and deadline={deadline}")
        object.__setattr__(self, "arrival", arrival)
        object.__setattr__(self, "deadline", deadline)
        object.__setattr__(self, "demand_id", check_int_range(
            demand_id, "demand_id", 0, MAX_ID))
        object.__setattr__(self, "customer_id", check_int_range(
            customer_id, "customer_id", 0, MAX_ID))
        object.__setattr__(self, "product_id", check_int_range(
            product_id, "product_id", 0, MAX_ID))
        object.__setattr__(self, "amount", check_int_range(
            amount, "amount", 1, MAX_ID))
        if not isinstance(measure, bool):
            raise type_error(measure, "measure", bool)
        object.__setattr__(self, "measure", measure)

    def __str__(self) -> str:
        """
        Get a string representation of the demand.

        :return: the string representation

        >>> str(Demand(arrival=16, deadline=28.0, demand_id=1,
        ...     customer_id=2, product_id=6, amount=12, measure=False))
        'd(id: 1, p: 6, c: 2, am: 12, ar: 16, dl: 28, me: F)'
        """
        fts: Final[Callable] = float_to_str
        return (f"d(id: {self.demand_id}, p: {self.product_id}, "
                f"c: {self.customer_id}, am: {self.amount}, "
                f"ar: {fts(self.arrival)}, dl: {fts(self.deadline)}, "
                f"me: {bool_to_str(self.measure)})")

    def __getitem__(self, item: int) -> int | float:
        """
        Access an element of this demand via an index.

        :param item: the index
        :return: the demand value at that index

        >>> d = Demand(arrival=16, deadline=28, demand_id=1,
        ...        customer_id=2, product_id=6, amount=12, measure=True)
        >>> d[0]
        1
        >>> d[1]
        2
        >>> d[2]
        6
        >>> d[3]
        12
        >>> d[4]
        16
        >>> d[5]
        28
        """
        if item == DEMAND_ID:
            return self.demand_id
        if item == DEMAND_CUSTOMER:
            return self.customer_id
        if item == DEMAND_PRODUCT:
            return self.product_id
        if item == DEMAND_AMOUNT:
            return self.amount
        if item == DEMAND_ARRIVAL:
            return try_int(self.arrival)
        if item == DEMAND_DEADLINE:
            return try_int(self.deadline)
        raise IndexError(
            f"index {item} out of bounds [0,{DEMAND_DEADLINE}].")

    def __iter__(self) -> Iterator[int | float]:
        """
        Iterate over the values in this demand.

        :return: the demand iterable

        >>> d = Demand(arrival=16, deadline=28, demand_id=1,
        ...        customer_id=2, product_id=6, amount=12, measure=True)
        >>> list(d)
        [1, 2, 6, 12, 16, 28]
        """
        yield self.demand_id  # DEMAND_ID
        yield self.customer_id  # DEMAND_CUSTOMER
        yield self.product_id  # DEMAND_PRODUCT:
        yield self.amount  # DEMAND_AMOUNT:
        yield try_int(self.arrival)  # DEMAND_ARRIVAL
        yield try_int(self.deadline)  # DEMAND_DEADLINE

    def __len__(self) -> int:
        """
        Get the length of the demand record.

        :returns `6`: always

        >>> len(Demand(arrival=16, deadline=28, demand_id=1,
        ...        customer_id=2, product_id=6, amount=12, measure=True))
        6
        """
        return 6


def __to_tuple(source: Iterable[int | float],
               cache: Callable, empty_ok: bool = False,
               type_var: type = int) -> tuple:
    """
    Convert an iterable of type integer to a tuple.

    :param source: the data source
    :param cache: the cache
    :param empty_ok: are empty tuples OK?
    :param type_var: the type variable
    :return: the tuple

    >>> ppl = repr_cache()
    >>> k1 = __to_tuple([1, 2, 3], ppl)
    >>> print(k1)
    (1, 2, 3)

    >>> __to_tuple({2}, ppl)
    (2,)

    >>> k2 = __to_tuple([1, 2, 3], ppl)
    >>> print(k2)
    (1, 2, 3)
    >>> k1 is k2
    True

    >>> k3 = __to_tuple([3.4, 2.3, 3.1], ppl, type_var=float)
    >>> print(k3)
    (3.4, 2.3, 3.1)
    >>> k1 is k3
    False

    >>> k4 = __to_tuple([3.4, 2.3, 3.1], ppl, type_var=float)
    >>> print(k4)
    (3.4, 2.3, 3.1)
    >>> k3 is k4
    True

    >>> try:
    ...     __to_tuple([], ppl)
    ... except Exception as e:
    ...     print(e)
    row has length 0.

    >>> __to_tuple([], ppl, empty_ok=True)
    ()

    >>> try:
    ...     __to_tuple([1, 2.0], ppl)
    ... except Exception as e:
    ...     print(e)
    row[1] should be an instance of int but is float, namely 2.0.

    >>> try:
    ...     __to_tuple([1.1, 2.0, 4, 3.4], ppl, type_var=float)
    ... except Exception as e:
    ...     print(e)
    row[2] should be an instance of float but is int, namely 4.
    """
    use_row = source if isinstance(source, tuple) else tuple(source)
    if (tuple.__len__(use_row) <= 0) and (not empty_ok):
        raise ValueError("row has length 0.")
    for j, v in enumerate(use_row):
        if not isinstance(v, type_var):
            raise type_error(v, f"row[{j}]", type_var)
        if not (isfinite(v) and (0 <= v <= MAX_VALUE)):  # type: ignore
            raise ValueError(f"row[{j}]={v} not in 0..{MAX_VALUE}")

    return cache(use_row)


def __to_npfloats(source: Iterable[int | float],  # pylint: disable=W1113
                  cache: Callable, empty_ok: bool = False,
                  *_) -> np.ndarray:  # pylint: disable=W1113
    """
    Convert to numpy floats.

    :param source: the source data
    :param cache: the cache
    :param empty_ok: are empty arrays OK?
    :return: the arrays

    >>> ppl = repr_cache()
    >>> a = __to_npfloats([3.4, 2.3, 3.1], ppl)
    >>> a
    array([3.4, 2.3, 3.1])
    >>> b = __to_npfloats([3.4, 2.3, 3.1], ppl)
    >>> b is a
    True

    >>> c = __to_npfloats([], ppl, empty_ok=True)
    >>> c
    array([], dtype=float64)

    >>> d = __to_npfloats([], ppl, empty_ok=True)
    >>> d is c
    True
    """
    return cache(np.array(__to_tuple(
        source, cache, empty_ok, float), np.float64))


def __to_nested_tuples(source: Iterable,
                       cache: Callable, empty_ok: bool = False,
                       type_var: type = int,
                       inner: Callable = __to_tuple) -> tuple:
    """
    Turn nested iterables of ints into nested tuples.

    :param source: the source list
    :param cache: the cache
    :param empty_ok: are empty tuples OK?
    :param type_var: the type variable
    :param inner: the inner function
    :return: the tuple or array

    >>> ppl = repr_cache()
    >>> k1 = __to_nested_tuples([(1, 2), [3, 2]], ppl)
    >>> print(k1)
    ((1, 2), (3, 2))

    >>> k2 = __to_nested_tuples([(1, 2), (1, 2, 4),  (1, 2)], ppl)
    >>> print(k2)
    ((1, 2), (1, 2, 4), (1, 2))

    >>> k2[0] is k2[2]
    True

    >>> k1[0] is k2[0]
    True
    >>> k1[0] is k2[2]
    True

    >>> __to_nested_tuples([(1, 2), (1, 2, 4),  (1, 2), []], ppl, True)
    ((1, 2), (1, 2, 4), (1, 2), ())

    >>> __to_nested_tuples([(), {}, []], ppl, True)
    ()

    >>> __to_nested_tuples([(1.0, 2.4), (1.0, 2.2)], ppl, True, float)
    ((1.0, 2.4), (1.0, 2.2))
    """
    if not isinstance(source, Iterable):
        raise type_error(source, "source", Iterable)
    dest: list = []
    ins: int = 0
    for row in source:
        use_row = inner(row, cache, empty_ok, type_var)
        ins += len(use_row)
        dest.append(use_row)

    if (ins <= 0) and empty_ok:  # if all inner tuples are empty,
        dest.clear()             # clear the tuple source

    n_rows: Final[int] = list.__len__(dest)
    if (n_rows <= 0) and (not empty_ok):
        raise ValueError("Got empty set of rows!")

    return cache(tuple(dest))


def __to_tuples(source: Iterable[Iterable],
                cache: Callable, empty_ok: bool = False, type_var=int,
                inner: Callable = __to_tuple) \
        -> tuple[tuple, ...]:
    """
    Turn 2D nested iterables into 2D nested tuples.

    :param source: the source
    :param cache: the cache
    :param empty_ok: are empty tuples OK?
    :param type_var: the type variable
    :param inner: the inner callable
    :return: the nested tuples

    >>> ppl = repr_cache()
    >>> k1 = __to_tuples([(1, 2), [3, 2]], ppl)
    >>> print(k1)
    ((1, 2), (3, 2))

    >>> k2 = __to_tuples([(1, 2), (1, 2, 4),  (1, 2)], ppl)
    >>> print(k2)
    ((1, 2), (1, 2, 4), (1, 2))

    >>> k2[0] is k2[2]
    True
    >>> k1[0] is k2[0]
    True
    >>> k1[0] is k2[2]
    True
    """
    return __to_nested_tuples(source, cache, empty_ok, type_var, inner)


def __to_2d_npfloat(source: Iterable[Iterable],  # pylint: disable=W1113
                    cache: Callable, empty_ok: bool = False,
                    *_) -> tuple[np.ndarray, ...]:  # pylint: disable=W1113
    """
    Turn 2D nested iterables into 2D nested tuples.

    :param source: the source
    :param cache: the cache
    :param empty_ok: are empty tuples OK?
    :param inner: the inner callable
    :return: the nested tuples

    >>> ppl = repr_cache()
    >>> k2 = __to_2d_npfloat([(1.0, 2.0), (1.0, 2.0, 4.0),  (1.0, 0.2)], ppl)
    >>> print(k2)
    (array([1., 2.]), array([1., 2., 4.]), array([1. , 0.2]))
    """
    return __to_nested_tuples(source, cache, empty_ok, float, __to_npfloats)


def __to_3d_npfloat(source: Iterable[Iterable[Iterable]],
                    cache: Callable, empty_ok: bool) \
        -> tuple[tuple[np.ndarray, ...], ...]:
    """
    Turn 3D nested iterables into 3D nested tuples.

    :param source: the source
    :param cache: the cache
    :param empty_ok: are empty tuples OK?
    :return: the nested tuples

    >>> ppl = repr_cache()
    >>> k1 = __to_3d_npfloat([[[3.0, 2.0], [44.0, 5.0], [2.0]],
    ...                        [[2.0], [5.0, 7.0]]], ppl, False)
    >>> print(k1)
    ((array([3., 2.]), array([44.,  5.]), array([2.])), \
(array([2.]), array([5., 7.])))
    >>> k1[0][2] is k1[1][0]
    True
    """
    return __to_nested_tuples(source, cache, empty_ok, float, __to_2d_npfloat)


def _make_routes(
        n_products: int, n_stations: int,
        source: Iterable[Iterable[int]],
        cache: Callable) -> tuple[tuple[int, ...], ...]:
    """
    Create the routes through stations for the products.

    Each product passes through a set of stations. It can pass through each
    station at most once. It can only pass through valid stations.

    :param n_products: the number of products
    :param n_stations: the number of stations
    :param source: the source data
    :param cache: the cache
    :return: the routes, a tuple of tuples

    >>> ppl = repr_cache()
    >>> _make_routes(2, 3, ((1, 2), (1, 0)), ppl)
    ((1, 2), (1, 0))

    >>> _make_routes(3, 3, ((1, 2), (1, 0), (0, 1, 2)), ppl)
    ((1, 2), (1, 0), (0, 1, 2))

    >>> k = _make_routes(3, 3, ((1, 2), (1, 2), (0, 1, 2)), ppl)
    >>> k[0] is k[1]
    True
    """
    check_int_range(n_products, "n_products", 1, MAX_ID)
    check_int_range(n_stations, "n_stations", 1, MAX_ID)
    dest: tuple[tuple[int, ...], ...] = __to_tuples(source, cache)

    n_rows: Final[int] = tuple.__len__(dest)
    if n_rows != n_products:
        raise ValueError(f"{n_products} products, but {n_rows} routes.")
    for i, route in enumerate(dest):
        stations: int = tuple.__len__(route)
        if stations <= 0:
            raise ValueError(
                f"len(row[{i}])={stations} but n_stations={n_stations}")
        for j, v in enumerate(route):
            if not 0 <= v < n_stations:
                raise ValueError(
                    f"row[{i},{j}]={v}, but n_stations={n_stations}")
    return dest


def __to_demand(
        source: Iterable[int | float], time_end_warmup: float,
        cache: Callable) -> Demand:
    """
    Convert an integer source to a tuple or a demand.

    :param source: the source
    :param time_end_warmup: the end of the warmup time
    :param cache: the cache
    :return: the Demand

    >>> ppl = repr_cache()
    >>> d1 = __to_demand([1, 2, 3, 20, 10, 100], 10.0, ppl)
    >>> d1
    Demand(arrival=10.0, deadline=100.0, demand_id=1, \
customer_id=2, product_id=3, amount=20, measure=True)
    >>> d2 = __to_demand([1, 2, 3, 20, 10, 100], 10.0, ppl)
    >>> d1 is d2
    True
    """
    if isinstance(source, Demand):
        return cast("Demand", source)
    tup: tuple[int | float, ...] = tuple(source)
    dl: int = tuple.__len__(tup)
    if dl != 6:
        raise ValueError(f"Expected 6 values, got {dl}.")
    arrival: int | float = tup[DEMAND_ARRIVAL]
    return cache(Demand(
        demand_id=cast("int", tup[DEMAND_ID]),
        customer_id=cast("int", tup[DEMAND_CUSTOMER]),
        product_id=cast("int", tup[DEMAND_PRODUCT]),
        amount=cast("int", tup[DEMAND_AMOUNT]),
        arrival=arrival, deadline=tup[DEMAND_DEADLINE],
        measure=time_end_warmup <= arrival))


def _make_demands(n_products: int, n_customers: int, n_demands: int,
                  source: Iterable[Iterable[int | float]],
                  time_end_warmup: float,
                  time_end_measure: float, cache: Callable) \
        -> tuple[Demand, ...]:
    """
    Create the demand records, sorted by release time.

    Each demand is a tuple of demand_id, customer_id, product_id, amount,
    release time, and deadline.

    :param n_products: the number of products
    :param n_customers: the number of customers
    :param n_demands: the number of demands
    :param time_end_warmup: the end of the warmup time
    :param time_end_measure: the end of the measure time period
    :param source: the source data
    :param cache: the cache
    :return: the demand tuples

    >>> ppl = repr_cache()
    >>> _make_demands(10, 10, 4, [[0, 2, 1, 4, 20, 21],
    ...     [2, 5, 2, 6, 17, 27],
    ...     [1, 6, 7, 12, 17, 21],
    ...     [3, 7, 3, 23, 5, 21]], 10.0, 1000.0, ppl)
    (Demand(arrival=5.0, deadline=21.0, demand_id=3, customer_id=7,\
 product_id=3, amount=23, measure=False),\
 Demand(arrival=17.0, deadline=21.0, demand_id=1, customer_id=6,\
 product_id=7, amount=12, measure=True),\
 Demand(arrival=17.0, deadline=27.0, demand_id=2, customer_id=5,\
 product_id=2, amount=6, measure=True),\
 Demand(arrival=20.0, deadline=21.0, demand_id=0, customer_id=2,\
 product_id=1, amount=4, measure=True))
    """
    check_int_range(n_products, "n_products", 1, MAX_ID)
    check_int_range(n_customers, "n_customers", 1, MAX_ID)
    check_int_range(n_demands, "n_demands", 1, MAX_ID)

    def __make_demand(ssss: Iterable[int | float],
                      ccc: Callable, *_) -> Demand:
        return __to_demand(ssss, time_end_warmup, ccc)

    temp: tuple[Demand, ...] = __to_nested_tuples(
        source, cache, False, inner=__make_demand)
    n_dem: int = tuple.__len__(temp)
    if n_dem != n_demands:
        raise ValueError(f"Expected {n_demands} demands, got {n_dem}?")

    used_ids: set[int] = set()
    min_id: int = 1000 * MAX_ID
    max_id: int = -1000 * MAX_ID
    dest: list[Demand] = []

    for i, demand in enumerate(temp):
        d_id: int = demand.demand_id
        if not 0 <= d_id < n_demands:
            raise ValueError(f"demand[{i}].id = {d_id}")
        if d_id in used_ids:
            raise ValueError(f"demand[{i}].id {d_id} appears twice!")
        used_ids.add(d_id)
        min_id = min(min_id, d_id)
        max_id = max(max_id, d_id)

        c_id: int = demand.customer_id
        if not 0 <= c_id < n_customers:
            raise ValueError(f"demand[{i}].customer = {c_id}, "
                             f"but n_customers={n_customers}")

        p_id: int = demand.product_id
        if not 0 <= p_id < n_products:
            raise ValueError(f"demand[{i}].product = {p_id}, "
                             f"but n_products={n_products}")

        amount: int = demand.amount
        if not 0 < amount < MAX_ID:
            raise ValueError(f"demand[{i}].amount = {amount}.")

        arrival: float = demand.arrival
        if not (isfinite(arrival) and (0 < arrival < MAX_ID)):
            raise ValueError(f"demand[{i}].arrival = {arrival}.")

        deadline: float = demand.deadline
        if not (isfinite(deadline) and arrival <= deadline < MAX_ID):
            raise ValueError(f"demand[{i}].deadline = {deadline}.")

        if arrival >= time_end_measure:
            raise ValueError(f"Demand[{i}]={demand!r} has arrival after "
                             "end of measurement period.")
        dest.append(demand)

    sl: int = set.__len__(used_ids)
    if sl != n_demands:
        raise ValueError(f"Got {n_demands} demands, but {sl} ids???")
    if ((max_id - min_id + 1) != n_demands) or (min_id != 0):
        raise ValueError(f"Invalid demand id range [{min_id}, {max_id}].")
    dest.sort()
    return cache(tuple(dest))


def _make_in_warehouse(n_products: int, source: Iterable[int],
                       cache: Callable) \
        -> tuple[int, ...]:
    """
    Make the amount of product in the warehouse at time 0.

    :param n_products: the total number of products
    :param source: the data source
    :param cache: the tuple cache
    :return: the amount of products in the warehouse

    >>> _make_in_warehouse(3, [1, 2, 3], repr_cache())
    (1, 2, 3)
    """
    ret: tuple[int, ...] = __to_tuple(source, cache)
    rl: Final[int] = tuple.__len__(ret)
    if rl != n_products:
        raise ValueError(f"We have {n_products} products, "
                         f"but the warehouse list length is {rl}.")
    for p, v in enumerate(ret):
        if not 0 <= v <= MAX_ID:
            raise ValueError(f"Got {v} units of product {p} in warehouse?")
    return ret


def _make_station_product_unit_times(
        n_products: int, n_stations: int,
        routes: tuple[tuple[float, ...], ...],
        source: Iterable[Iterable[Iterable[float]]],
        cache: Callable) -> tuple[tuple[np.ndarray, ...], ...]:
    """
    Create the structure for the work times per product unit per station.

    Here we have for each station, for each product, a sequence of per-unit
    production settings. Each such "production settings" is a tuple with a
    per-unit production time and an end time index until which it is valid.
    Production times cycle, so if we produce something after the last end
    time index, we begin again at production time index 0.

    :param n_products: the number of products
    :param n_stations: the number of stations
    :param routes: the routes of the products through the stations
    :param source: the source array
    :param cache: the cache
    :return: the station unit times

    >>> ppl = repr_cache()
    >>> rts = _make_routes(3, 2, [[0, 1], [0], [1, 0]], ppl)
    >>> print(rts)
    ((0, 1), (0,), (1, 0))

    >>> mpt1 = _make_station_product_unit_times(3, 2, rts, [
    ...     [[1.0, 2.0, 3.0, 5.0], [1.0, 2.0, 3.0, 5.0],
    ...      [1.0, 10.0, 2.0, 30.0]],
    ...     [[2.0, 20.0, 3.0, 40.0], [], [4.0, 56.0, 34.0, 444.0]]], ppl)
    >>> print(mpt1)
    ((array([1., 2., 3., 5.]), array([1., 2., 3., 5.]), \
array([ 1., 10.,  2., 30.])), (array([ 2., 20.,  3., 40.]), \
array([], dtype=float64), array([  4.,  56.,  34., 444.])))
    >>> mpt1[0][0] is mpt1[0][1]
    True

    >>> mpt2 = _make_station_product_unit_times(3, 2, rts, [
    ...     [[1.0, 2.0, 3.0, 5.0], [1.0, 2.0, 3.0, 5.0],
    ...      [1.0, 10.0, 2.0, 30.0]],
    ...     [[2.0, 20.0, 3.0, 40.0], [], [4.0, 56.0, 34.0, 444.0]]], ppl)
    >>> print(mpt2)
    ((array([1., 2., 3., 5.]), array([1., 2., 3., 5.]), \
array([ 1., 10.,  2., 30.])), (array([ 2., 20.,  3., 40.]), \
array([], dtype=float64), array([  4.,  56.,  34., 444.])))
    >>> mpt1 is mpt2
    True
    """
    ret: tuple[tuple[np.ndarray, ...], ...] = __to_3d_npfloat(
        source, cache, True)

    if tuple.__len__(routes) != n_products:
        raise ValueError("invalid routes!")

    d1: int = tuple.__len__(ret)
    if d1 != n_stations:
        raise ValueError(
            f"Got {d1} station-times, but {n_stations} stations.")
    for mid, station in enumerate(ret):
        d2: int = tuple.__len__(station)
        if d2 <= 0:
            for pid, r in enumerate(routes):
                if mid in r:
                    raise ValueError(
                        f"Station {mid} in route for product {pid}, "
                        "but has no production time")
            continue
        if d2 != n_products:
            raise ValueError(f"got {d2} products for station {mid}, "
                             f"but have {n_products} products")
        for pid, product in enumerate(station):
            needs_times: bool = mid in routes[pid]
            d3: int = np.ndarray.__len__(product)
            if (not needs_times) and (d3 > 0):
                raise ValueError(
                    f"product {pid} does not pass through station {mid}, "
                    "so there must not be production times!")
            if needs_times and (d3 <= 0):
                raise ValueError(
                    f"product {pid} does pass through station {mid}, "
                    "so there must be production times!")
            if (d3 % 2) != 0:
                raise ValueError(
                    f"production times for {pid} does pass through station "
                    f"{mid}, must be of even length, but got length {d3}.")
            last_end = 0
            for pt, time in enumerate(batched(product, 2)):
                if tuple.__len__(time) != 2:
                    raise ValueError(f"production times must be 2-tuples, "
                                     f"but got {time} for product {pid} on "
                                     f"station {mid} at position {pt}")
                unit_time, end = time
                if not ((unit_time > 0) and (last_end < end < MAX_ID)):
                    raise ValueError(
                        f"Invalid unit time {unit_time} and end time "
                        f"{end} for product {pid} on station {mid}")
                last_end = end

    return ret


def _make_infos(source: Iterable[tuple[str, str]] | Mapping[str, str] | None)\
        -> Mapping[str, str]:
    """
    Make the additional information record.

    :param source: the information to represent
    :return: the information record
    """
    use_source: Iterable[tuple[str, str]] = () if source is None else (
        source.items() if isinstance(source, Mapping) else source)
    if not isinstance(use_source, Iterable):
        raise type_error(source, "infos", Iterable)
    dst: dict[str, str] = {}
    for i, tup in enumerate(use_source):
        if tuple.__len__(tup) != 2:
            raise ValueError(f"Invalid tuple {tup} at index {i} in infos.")
        k: str = str.strip(tup[0])
        v: str = str.strip(tup[1])
        if (str.__len__(k) <= 0) or (str.__len__(v) <= 0):
            raise ValueError(f"Invalid key/values {k!r}/{v!r} in tuple "
                             f"{tup} at index {i} in infos.")
        if __FORBIDDEN_INFO_KEYS(str.lower(k)):
            raise ValueError(
                f"Info key {k!r} in tuple {tup} forbidden at index {i}.")
        if not all(map(__ALLOWED_INFO_KEY_CHARS, k)):
            raise ValueError(
                f"Malformed info key {k!r} in tuple {tup} at index {i}.")
        if k in dst:
            raise ValueError(f"Duplicate key {k!r} found in tuple {tup} "
                             f"at index {i} in infos.")
        dst[k] = v
    return immutable_mapping(dst)


class Instance(Component):
    """An instance of the Production Scheduling Problem."""

    def __init__(
            self, name: str,
            n_products: int, n_customers: int, n_stations: int,
            n_demands: int,
            time_end_warmup: int | float, time_end_measure: int | float,
            routes: Iterable[Iterable[int]],
            demands: Iterable[Iterable[int | float]],
            warehous_at_t0: Iterable[int],
            station_product_unit_times: Iterable[Iterable[Iterable[float]]],
            infos: Iterable[tuple[str, str]] | Mapping[
                str, str] | None = None) \
            -> None:
        """
        Create an instance of the production scheduling time.

        :param name: the instance name
        :param n_products: the number of products
        :param n_customers: the number of customers
        :param n_stations: the number of stations
        :param n_demands: the number of demand records
        :param time_end_warmup: the time unit when the warmup time ends and the
            actual measurement begins
        :param time_end_measure: the time unit when the actual measure time
            ends
        :param routes: for each product, the sequence of stations that it has
            to pass
        :param demands: a sequences of demands of the form (
            customer_id, product_id, product_amount, release_time) OR a
            sequence of :class:`Demand` records.
        :param warehous_at_t0: the amount of products in the warehouse at time
            0 for each product
        :param station_product_unit_times: for each station and each product
            the per-unit-production time schedule, in the form of
            "per_unit_time, duration", where duration is the number of time
            units for which the per_unit_time is value
        :param station_product_unit_times: the cycling unit times for each
            product on each station, each with a validity duration
        :param infos: additional infos to be stored with the instance.
            These are key-value pairs with keys that are not used by the
            instance. They have no impact on the instance performance, but may
            explain settings of an instance generator.
        """
        use_name: Final[str] = sanitize_name(name)
        if name != use_name:
            raise ValueError(f"Name {name!r} is not a valid name.")
        if not all(map(_ALLOWED_NAME_CHARS, name)):
            raise ValueError(f"Name {name!r} contains invalid characters.")
        #: the name of this instance
        self.name: Final[str] = name

        #: the number of products in the scenario
        self.n_products: Final[int] = check_int_range(
            n_products, "n_products", 1, MAX_ID)
        #: the number of customers in the scenario
        self.n_customers: Final[int] = check_int_range(
            n_customers, "n_customers", 1, MAX_ID)
        #: the number of stations or workstations in the scenario
        self.n_stations: Final[int] = check_int_range(
            n_stations, "n_stations", 1, MAX_ID)
        #: the number of demands in the scenario
        self.n_demands: Final[int] = check_int_range(
            n_demands, "n_demands", 1, MAX_ID)

        if not isinstance(time_end_warmup, int | float):
            raise type_error(time_end_warmup, "time_end_warmup", (int, float))
        time_end_warmup = float(time_end_warmup)
        if not (isfinite(time_end_warmup) and (
                0 <= time_end_warmup < MAX_VALUE)):
            raise ValueError(f"Invalid time_end_warmup={time_end_warmup}.")
        #: the end of the warmup time
        self.time_end_warmup: Final[float] = time_end_warmup

        if not isinstance(time_end_measure, int | float):
            raise type_error(time_end_measure, "time_end_measure", (
                int, float))
        time_end_measure = float(time_end_measure)
        if not (isfinite(time_end_measure) and (
                time_end_warmup < time_end_measure < MAX_VALUE)):
            raise ValueError(f"Invalid time_end_measure={time_end_measure} "
                             f"for time_end_warmup={time_end_warmup}.")
        #: the end of the measurement time
        self.time_end_measure: Final[float] = time_end_measure

        cache: Final[Callable] = repr_cache()  # the pool for resolving tuples

        #: the product routes, i.e., the stations through which each product
        #: must pass
        self.routes: Final[tuple[tuple[int, ...], ...]] = _make_routes(
            n_products, n_stations, routes, cache)
        #: The demands: Each demand is a tuple of demand_id, customer_id,
        #: product_id, amount, release_time, and deadline.
        #: The customer makes their order at time step release_time.
        #: They expect to receive their product by the deadline.
        #: The demands are sorted by release time and then deadline.
        #: The release time is always > 0.
        #: The deadline is always >= release time.
        #: Demand ids are unique.
        self.demands: Final[tuple[Demand, ...]] = _make_demands(
            n_products, n_customers, n_demands, demands, time_end_warmup,
            time_end_measure, cache)

        # count the demands that fall in the measure time window
        n_measure: int = 0
        n_measures: list[int] = [0] * n_products
        for d in self.demands:
            if d.arrival >= self.time_end_measure:
                raise ValueError(f"Invalid arrival time of demand {d!r}.")
            if d.measure != (self.time_end_warmup <= d.arrival):
                raise ValueError(
                    f"Inconsistent measure property for demand {d!r}.")
            if d.measure:
                n_measure += 1
                n_measures[d.product_id] += 1
        if n_measure <= 0:
            raise ValueError("There are no measurable demands!")
        for pid, npm in enumerate(n_measures):
            if npm <= 0:
                raise ValueError(f"No measurable demand for product {pid}!")
        #: the number of demands that actually fall into the time measured
        #: window
        self.n_measurable_demands: Final[int] = n_measure
        #: the measurable demands on a per-product basis
        self.n_measurable_demands_per_product: Final[tuple[int, ...]] = tuple(
            n_measures)

        #: The units of product in the warehouse at time step 0.
        #: For each product, we have either 0 or a positive amount of product.
        self.warehous_at_t0: Final[tuple[int, ...]] = _make_in_warehouse(
            n_products, warehous_at_t0, cache)

        #: The per-station unit production times for each product.
        #: Each station can have different production times per product.
        #: Let's say that this is tuple `A`.
        #: For each product, it has a tuple `B` at the index of the product
        #: id.
        #: If the product does not pass through the station, `B` is empty.
        #: Otherwise, it holds one or multiple tuples `C`.
        #: Each tuple `C` consists of two numbers:
        #: A per-unit-production time for the product.
        #: An end time index for this production time.
        #: Once the real time surpasses the end time of the last of these
        #: production specs, the production specs are recycled and begin
        #: again.
        self.station_product_unit_times: Final[tuple[tuple[
            np.ndarray, ...], ...]] = _make_station_product_unit_times(
            n_products, n_stations, self.routes, station_product_unit_times,
            cache)

        #: Additional information about the nature of the instance can be
        #: stored here. This has no impact on the behavior of the instance,
        #: but it may explain, e.g., settings of an instance generator.
        self.infos: Final[Mapping[str, str]] = _make_infos(infos)

    def __str__(self):
        """
        Get the name of this instance.

        :return: the name of this instance
        """
        return self.name

    def _tuple(self) -> tuple:
        """
        Convert this object to a tuple.

        :return: the tuple

        >>> Instance(name="test1", n_products=1, n_customers=1, n_stations=2,
        ...         n_demands=1, time_end_warmup=12, time_end_measure=30,
        ...         routes=[[0, 1]], demands=[[0, 0, 0, 10, 20, 100]],
        ...         warehous_at_t0=[0],
        ...     station_product_unit_times=[[[10.0, 10000.0]],
        ...                                 [[30.0, 10000.0]]])._tuple()
        ('test1', 2, 1, 1, 1, 12.0, 30.0, (Demand(arrival=20.0,\
 deadline=100.0, demand_id=0, customer_id=0, product_id=0, amount=10,\
 measure=True),), ((0, 1),), (0,), (), ((10.0, 10000.0), (30.0, 10000.0)))
        """
        return (self.name, self.n_stations, self.n_products,
                self.n_demands, self.n_customers, self.time_end_warmup,
                self.time_end_measure, self.demands,
                self.routes, self.warehous_at_t0, tuple(self.infos.items()),
                tuple(tuple(float(x) for x in a2) for a1 in
                      self.station_product_unit_times for a2 in a1))

    def __eq__(self, other):
        """
        Compare this object with another object.

        :param other: the other object
        :return: `NotImplemented` if the other object is not an `Instance`,
            otherwise the equality comparison result.

        >>> i1 = Instance(name="test1", n_products=1, n_customers=1,
        ...         n_stations=2, n_demands=1,
        ...         time_end_warmup=12, time_end_measure=30,
        ...         routes=[[0, 1]],
        ...         demands=[[0, 0, 0, 10, 20, 100]],
        ...         warehous_at_t0=[0],
        ...     station_product_unit_times=[[[10.0, 10000.0]],
        ...                                 [[30.0, 10000.0]]])
        >>> i2 = Instance(name="test1", n_products=1, n_customers=1,
        ...         n_stations=2, n_demands=1,
        ...         time_end_warmup=12, time_end_measure=30,
        ...         routes=[[0, 1]],
        ...         demands=[[0, 0, 0, 10, 20, 100]],
        ...         warehous_at_t0=[0],
        ...     station_product_unit_times=[[[10.0, 10000.0]],
        ...                                 [[30.0, 10000.0]]])
        >>> i1 == i2
        True
        >>> i3 = Instance(name="test1", n_products=1, n_customers=1,
        ...         n_stations=2, n_demands=1,
        ...         time_end_warmup=12, time_end_measure=30,
        ...         routes=[[0, 1]],
        ...         demands=[[0, 0, 0, 10, 20, 100]],
        ...         warehous_at_t0=[0],
        ...     station_product_unit_times=[[[10.0, 10000.1]],
        ...                                 [[30.0, 10000.0]]])
        >>> i1 == i3
        False
        """
        if other is None:
            return False
        if not isinstance(other, Instance):
            return NotImplemented
        return self._tuple() == cast("Instance", other)._tuple()

    def __hash__(self) -> int:
        """
        Get the hash code of this object.

        :return: the hash code of this object
        """
        return hash(self._tuple())


#: the instance name key
KEY_NAME: Final[str] = "name"
#: the key for the number of products
KEY_N_PRODUCTS: Final[str] = "n_products"
#: the key for the number of customers
KEY_N_CUSTOMERS: Final[str] = "n_customers"
#: the key for the number of stations
KEY_N_STATIONS: Final[str] = "n_stations"
#: the number of demands in the scenario
KEY_N_DEMANDS: Final[str] = "n_demands"
#: the end of the warmup period
KEY_TIME_END_WARMUP: Final[str] = "time_end_warmup"
#: the end of the measure period
KEY_TIME_END_MEASURE: Final[str] = "time_end_measure"
#: the start of a key index
KEY_IDX_START: Final[str] = "["
#: the end of a key index
KEY_IDX_END: Final[str] = "]"
#: the first part of the product route key
KEY_ROUTE: Final[str] = "product_route"
#: the first part of the demand key
KEY_DEMAND: Final[str] = "demand"
#: The amount of products in the warehouse at time step 0.
KEY_IN_WAREHOUSE: Final[str] = "products_in_warehouse_at_t0"
#: the first part of the production time
KEY_PRODUCTION_TIME: Final[str] = "production_time"

#: the key value split string
_KEY_VALUE_SPLIT: Final[str] = str.strip(KEY_VALUE_SEPARATOR)

#: the forbidden keys
__FORBIDDEN_INFO_KEYS: Final[Callable[[str], bool]] = {
    KEY_NAME, KEY_N_PRODUCTS, KEY_N_CUSTOMERS, KEY_N_STATIONS,
    KEY_N_DEMANDS, KEY_TIME_END_MEASURE, KEY_TIME_END_WARMUP,
    KEY_ROUTE, KEY_DEMAND, KEY_IN_WAREHOUSE, KEY_PRODUCTION_TIME}.__contains__

#: the allowed information key characters
__ALLOWED_INFO_KEY_CHARS: Final[Callable[[str], bool]] = set(
    ascii_letters + digits + "_." + KEY_IDX_START + KEY_IDX_END).__contains__

#: the allowed characters in names
_ALLOWED_NAME_CHARS: Final[Callable[[str], bool]] = set(
    ascii_letters + digits + "_").__contains__


def to_stream(instance: Instance) -> Generator[str, None, None]:
    """
    Convert an instance to a stream of data.

    :param instance: the instance to convert to a stream
    :return: the stream of data
    """
    if not isinstance(instance, Instance):
        raise type_error(instance, "instance", Instance)

    yield f"{COMMENT_START} --- the data of instance {instance.name!r} ---"
    yield COMMENT_START
    yield (f"{COMMENT_START} Lines beginning with {COMMENT_START!r} are "
           f"comments.")
    yield COMMENT_START
    yield f"{COMMENT_START} the unique identifying name of the instance"
    yield f"{KEY_NAME}{KEY_VALUE_SEPARATOR}{instance.name}"
    yield COMMENT_START
    yield f"{COMMENT_START} the number of products in the instance, > 0"
    yield f"{KEY_N_PRODUCTS}{KEY_VALUE_SEPARATOR}{instance.n_products}"
    yield (f"{COMMENT_START} Valid product indices are in 0.."
           f"{instance.n_products - 1}.")
    yield COMMENT_START
    yield f"{COMMENT_START} the number of customers in the instance, > 0"
    yield f"{KEY_N_CUSTOMERS}{KEY_VALUE_SEPARATOR}{instance.n_customers}"
    yield (f"{COMMENT_START} Valid customer indices are in 0.."
           f"{instance.n_customers - 1}.")
    yield COMMENT_START
    yield f"{COMMENT_START} the number of stations in the instance, > 0"
    yield f"{KEY_N_STATIONS}{KEY_VALUE_SEPARATOR}{instance.n_stations}"
    yield (f"{COMMENT_START} Valid station indices are in 0.."
           f"{instance.n_stations - 1}.")
    yield COMMENT_START
    yield (f"{COMMENT_START} the number of customer orders (demands) issued "
           f"by the customers, > 0")
    yield f"{KEY_N_DEMANDS}{KEY_VALUE_SEPARATOR}{instance.n_demands}"
    yield (f"{COMMENT_START} Valid demand/order indices are in 0.."
           f"{instance.n_demands - 1}.")
    yield COMMENT_START
    yield (f"{COMMENT_START} end of the warmup period in the simulations, "
           f">= 0")
    wm: Final[str] = float_to_str(instance.time_end_warmup)
    yield f"{KEY_TIME_END_WARMUP}{KEY_VALUE_SEPARATOR}{wm}"
    yield (f"{COMMENT_START} The simulation will not measure anything "
           f"during the first {wm} time units.")
    yield COMMENT_START
    yield (f"{COMMENT_START} end of the measurement period in the "
           f"simulations, > {wm}")
    meas: Final[str] = float_to_str(instance.time_end_measure)
    yield f"{KEY_TIME_END_MEASURE}{KEY_VALUE_SEPARATOR}{meas}"
    yield (f"{COMMENT_START} The simulation will only measure things during "
           f" left-closed and right-open interval [{wm},{meas}).")

    yield COMMENT_START
    yield (f"{COMMENT_START} For each product, we now specify the indices of "
           f"the stations by which it will be processed, in the order in "
           f"which it will be processed by them.")
    yield (f"{COMMENT_START} {KEY_ROUTE}{KEY_IDX_START}0"
           f"{KEY_IDX_END} is the production route by the first product, "
           "which has index 0.")
    route_0: tuple[int, ...] = instance.routes[0]
    yield (f"{COMMENT_START} This product is processed by "
           f"{tuple.__len__(route_0)} stations, namely first by the "
           f"station with index {int(route_0[0])} and last by the station "
           f"with index {int(route_0[-1])}.")
    for p, route in enumerate(instance.routes):
        yield (f"{KEY_ROUTE}{KEY_IDX_START}{p}{KEY_IDX_END}"
               f"{KEY_VALUE_SEPARATOR}"
               f"{CSV_SEPARATOR.join(map(str, route))}")

    yield COMMENT_START
    yield (f"{COMMENT_START} For each customer order/demand, we now "
           f"specify the following values:")
    yield f"{COMMENT_START} 1. the demand ID in square brackets"
    yield f"{COMMENT_START} 2. the ID of the customer who made the order"
    yield f"{COMMENT_START} 3. the ID of the product that the customer ordered"
    yield (f"{COMMENT_START} 4. the amount of the product that the customer"
           " ordered")
    yield (f"{COMMENT_START} 5. the arrival time of the demand, > 0, i.e., "
           f"the moment in time when the customer informed us that they want "
           f"the product")
    yield (f"{COMMENT_START} 6. the deadline, i.e., when the customer expects "
           f"the product, >= arrival time")
    srt: list[Demand] = sorted(instance.demands, key=lambda d: d.demand_id)
    fd: Demand = srt[0]
    yield (f"{COMMENT_START} For example, the demand with ID {fd.demand_id} "
           f"was issued by the customer with ID {fd.customer_id} for "
           f"{fd.amount} units of the product with ID "
           f"{fd.product_id}.")
    yield (f"{COMMENT_START} The order comes into the "
           f"system at time unit {fd.arrival} and the customer expects "
           f"the product to be ready at time unit {fd.deadline}.")
    for demand in srt:
        it = iter(demand)
        next(it)  # pylint: disable=R1708
        row: str = CSV_SEPARATOR.join(map(num_to_str, it))
        yield (f"{KEY_DEMAND}{KEY_IDX_START}{demand.demand_id}{KEY_IDX_END}"
               f"{KEY_VALUE_SEPARATOR}"
               f"{row}")

    yield COMMENT_START
    yield (f"{COMMENT_START} For each product, we now specify the amount "
           f"that is in the warehouse at time step 0.")
    yield (f"{COMMENT_START} For example, there are "
           f"{instance.warehous_at_t0[0]} units of product 0 in the "
           f"warehouse at the beginning of the simulation.")
    yield (f"{KEY_IN_WAREHOUSE}{KEY_VALUE_SEPARATOR}"
           f"{CSV_SEPARATOR.join(map(str, instance.warehous_at_t0))}")

    yield COMMENT_START
    yield (f"{COMMENT_START} For each station, we now specify the production "
           f"times for each product that passes through the station.")
    empty_pdx: tuple[int, int] | None = None
    filled_pdx: tuple[int, int, np.ndarray] | None = None
    need: int = 2
    for mid, station in enumerate(instance.station_product_unit_times):
        for pid, product in enumerate(station):
            pdl: int = np.ndarray.__len__(product)
            if (pdl <= 0) and (empty_pdx is None):
                empty_pdx = mid, pid
                need -= 1
                if need <= 0:
                    break
            elif (pdl > 0) and (filled_pdx is None):
                filled_pdx = mid, pid, product
                need -= 1
                if need <= 0:
                    break
        if need <= 0:
            break
    if empty_pdx is not None:
        yield (f"{COMMENT_START} For example, product {empty_pdx[1]} does "
               f"not pass through station {empty_pdx[0]}, so it is not "
               "listed here.")
    if filled_pdx is not None:
        yield (f"{COMMENT_START} For example, one unit of product "
               f"{filled_pdx[1]} passes through station {filled_pdx[0]}.")
        yield (f"{COMMENT_START} There, it needs {filled_pdx[2][0]} time "
               f"units per product unit from t=0 to t={filled_pdx[2][1]}.")
        if np.ndarray.__len__(filled_pdx[2]) > 2:
            yield (f"{COMMENT_START} After that, it needs {filled_pdx[2][2]}"
                   " time units per product unit until t="
                   f"{filled_pdx[2][3]}.")

    cache: dict[str, str] = {}
    for mid, station in enumerate(instance.station_product_unit_times):
        for pid, product in enumerate(station):
            if np.ndarray.__len__(product) <= 0:
                continue
            value: str = CSV_SEPARATOR.join(map(num_to_str, map(
                try_int, map(float, product))))
            key: str = (f"{KEY_PRODUCTION_TIME}{KEY_IDX_START}{mid}"
                        f"{CSV_SEPARATOR}{pid}{KEY_IDX_END}")
            if value in cache:
                value = cache[value]
            else:
                cache[value] = key
            yield f"{key}{KEY_VALUE_SEPARATOR}{value}"

    n_infos: Final[int] = len(instance.infos)
    if n_infos > 0:
        yield COMMENT_START
        yield (f"{COMMENT_START} The following {n_infos} key/value pairs "
               "denote additional information about the instance.")
        yield (f"{COMMENT_START} They have no impact whatsoever on the "
               "instance behavior.")
        yield (f"{COMMENT_START} A common use case is that we may have "
               "used a method to randomly sample the instance.")
        yield (f"{COMMENT_START} In this case, we could store the parameters "
               f"of the instance generator, such as the random seed and/or "
               f"the distributions used in this section.")
        for k, v in instance.infos.items():
            yield f"{k}{KEY_VALUE_SEPARATOR}{v}"


def __get_key_index(full_key: str) -> str:
    """
    Extract the key index from a key.

    :param full_key: the full key
    :return: the key index

    >>> __get_key_index("s[12 ]")
    '12'
    """
    start: int = str.index(full_key, KEY_IDX_START)
    end: int = str.index(full_key, KEY_IDX_END, start)
    if not 0 < start < end < str.__len__(full_key):
        raise ValueError(f"Invalid key {full_key!r}.")
    idx: str = str.strip(full_key[start + 1:end])
    if str.__len__(idx) <= 0:
        raise ValueError(f"Invalid index in key {full_key!r}.")
    return idx


def __pe(message: str, oline: str, line_idx: int) -> ValueError:
    """
    Create a value error to be raised inside the parser.

    :param message: the message
    :param oline: the original line
    :param line_idx: the line index
    :return: the error
    """
    return ValueError(f"{message} at line {line_idx + 1} ({oline!r})")


def from_stream(stream: Iterable[str]) -> Instance:
    """
    Read an instance from a data stream.

    :param stream: the data stream
    :return: the instance
    """
    if not isinstance(stream, Iterable):
        raise type_error(stream, "stream", Iterable)
    name: str | None = None
    n_products: int | None = None
    n_customers: int | None = None
    n_stations: int | None = None
    n_demands: int | None = None
    time_end_warmup: int | float | None = None
    time_end_measure: int | float | None = None
    routes: list[list[int]] | None = None
    demands: list[list[int | float]] | None = None
    in_warehouse: list[int] | None = None
    station_product_times: list[list[list[float]]] | None = None
    infos: dict[str, str] = {}

    for line_idx, oline in enumerate(stream):
        line = str.strip(oline)
        if str.startswith(line, COMMENT_START):
            continue

        split_idx: int = str.find(line, _KEY_VALUE_SPLIT)
        if split_idx > -1:
            key: str = str.lower(str.strip(line[:split_idx]))
            value: str = str.strip(
                line[split_idx + str.__len__(_KEY_VALUE_SPLIT):])
            if (str.__len__(key) <= 0) or (str.__len__(value) <= 0):
                raise __pe(f"Invalid key/value pair {key!r}/{value!r}",
                           oline, line_idx)

            if key == KEY_NAME:
                if name is not None:
                    raise __pe(f"{KEY_NAME} already defined as {name!r}, "
                               f"cannot be set to {value!r}", oline, line_idx)
                name = value

            elif key == KEY_N_STATIONS:
                if n_stations is not None:
                    raise __pe(
                        f"{KEY_N_STATIONS} already defined as {n_stations!r},"
                        f" cannot be set to {value!r}", oline, line_idx)
                n_stations = check_to_int_range(
                    value, KEY_N_STATIONS, 1, 1_000_000)

            elif key == KEY_N_PRODUCTS:
                if n_products is not None:
                    raise __pe(
                        f"{KEY_N_PRODUCTS} already defined as {n_products!r},"
                        f" cannot be set to {value!r}", oline, line_idx)
                n_products = check_to_int_range(
                    value, KEY_N_PRODUCTS, 1, 1_000_000)

            elif key == KEY_N_CUSTOMERS:
                if n_customers is not None:
                    raise __pe(
                        f"{KEY_N_CUSTOMERS} already defined as "
                        f"{n_customers!r}, cannot be set to {value!r}",
                        oline, line_idx)
                n_customers = check_to_int_range(
                    value, KEY_N_CUSTOMERS, 1, 1_000_000)

            elif key == KEY_N_DEMANDS:
                if n_demands is not None:
                    raise __pe(
                        f"{KEY_N_DEMANDS} already defined as {n_demands!r}, "
                        f"cannot be set to {value!r}", oline, line_idx)
                n_demands = check_to_int_range(
                    value, KEY_N_DEMANDS, 1, 1_000_000)

            elif key == KEY_TIME_END_WARMUP:
                if time_end_warmup is not None:
                    raise __pe(f"{KEY_TIME_END_WARMUP} already defined",
                               oline, line_idx)
                time_end_warmup = float(value)
                if not (isfinite(time_end_warmup) and (time_end_warmup > 0)):
                    raise __pe(f"time_end_warmup={time_end_warmup} invalid",
                               oline, line_idx)

            elif key == KEY_TIME_END_MEASURE:
                if time_end_measure is not None:
                    raise __pe(f"{KEY_TIME_END_MEASURE} already defined",
                               oline, line_idx)
                time_end_measure = float(value)
                if not (isfinite(time_end_measure) and (
                        time_end_measure > 0)):
                    raise __pe(f"time_end_measure={time_end_measure} invalid",
                               oline, line_idx)
                if (time_end_warmup is not None) and (
                        time_end_measure <= time_end_warmup):
                    raise __pe(f"time_end_warmup={time_end_warmup} and "
                               f"time_end_measure={time_end_measure}",
                               oline, line_idx)

            elif key == KEY_IN_WAREHOUSE:
                if in_warehouse is not None:
                    raise __pe(f"{KEY_IN_WAREHOUSE} already defined",
                               oline, line_idx)
                if n_products is None:
                    raise __pe(f"Must define {KEY_N_PRODUCTS} before "
                               f"{KEY_IN_WAREHOUSE}.", oline, line_idx)
                in_warehouse = list(map(int, str.split(
                    value, CSV_SEPARATOR)))
                if list.__len__(in_warehouse) != n_products:
                    raise __pe(
                        f"Expected {n_products} products in warehouse, got "
                        f"{in_warehouse}.", oline, line_idx)

            elif str.startswith(key, KEY_ROUTE):
                if n_products is None:
                    raise __pe(f"Must define {KEY_N_PRODUCTS} before "
                               f"{KEY_ROUTE}.", oline, line_idx)
                if n_stations is None:
                    raise ValueError(f"Must define {KEY_N_STATIONS} before "
                                     f"{KEY_ROUTE}.", oline, line_idx)
                if routes is None:
                    routes = [[] for _ in range(n_products)]

                product_id: int = check_to_int_range(
                    __get_key_index(key), KEY_ROUTE, 0, n_products - 1)
                rlst: list[int] = routes[product_id]
                if list.__len__(rlst) != 0:
                    raise __pe(
                        f"Already gave {KEY_ROUTE}{KEY_IDX_START}{product_id}"
                        f"{KEY_IDX_END}", oline, line_idx)
                rlst.extend(map(int, str.split(value, CSV_SEPARATOR)))
                if list.__len__(rlst) <= 0:
                    raise __pe(f"Route for product {product_id} is empty",
                               oline, line_idx)

            elif str.startswith(key, KEY_DEMAND):
                if n_customers is None:
                    raise __pe(f"Must define {KEY_N_CUSTOMERS} before "
                               f"{KEY_DEMAND}", oline, line_idx)
                if n_products is None:
                    raise __pe(f"Must define {KEY_N_PRODUCTS} before "
                               f"{KEY_DEMAND}.", oline, line_idx)
                if n_demands is None:
                    raise __pe(f"Must define {KEY_N_DEMANDS} before "
                               f"{KEY_DEMAND}", oline, line_idx)
                if demands is None:
                    demands = [[i] for i in range(n_demands)]

                demand_id: int = check_to_int_range(
                    __get_key_index(key), KEY_DEMAND, 0, n_demands - 1)
                dlst: list[int | float] = demands[demand_id]
                if list.__len__(dlst) != 1:
                    raise __pe(f"Already gave {KEY_DEMAND}{KEY_IDX_START}"
                               f"{demand_id}{KEY_IDX_END}", oline, line_idx)
                str_lst = str.split(value, CSV_SEPARATOR)
                if list.__len__(str_lst) != 5:
                    raise __pe(
                        f"Demand {demand_id} must have 5 entries, but got: "
                        f"{str_lst!r}", oline, line_idx)
                dlst.extend((int(str_lst[0]), int(str_lst[1]),
                             int(str_lst[2]), float(str_lst[3]),
                             float(str_lst[4])))

            elif str.startswith(key, KEY_PRODUCTION_TIME):
                if n_products is None:
                    raise __pe(f"Must define {KEY_N_PRODUCTS} before "
                               f"{KEY_PRODUCTION_TIME}", oline, line_idx)
                if n_stations is None:
                    raise __pe(f"Must define {KEY_N_STATIONS} before"
                               f" {KEY_PRODUCTION_TIME}", oline, line_idx)
                station, product = str.split(
                    __get_key_index(key), CSV_SEPARATOR)
                station_id: int = check_to_int_range(
                    station, "station", 0, n_stations - 1)
                product_id = check_to_int_range(
                    product, "product", 0, n_products - 1)

                if station_product_times is None:
                    station_product_times = \
                        [[[] for _ in range(n_products)]
                         for __ in range(n_stations)]

                if str.startswith(value, KEY_PRODUCTION_TIME):
                    station, product = str.split(
                        __get_key_index(value), CSV_SEPARATOR)
                    use_station_id: int = check_to_int_range(
                        station, "station", 0, n_stations - 1)
                    use_product_id = check_to_int_range(
                        product, "product", 0, n_products - 1)
                    station_product_times[station_id][product_id] = (
                        station_product_times)[use_station_id][use_product_id]
                else:
                    mpd: list[float] = station_product_times[
                        station_id][product_id]
                    if list.__len__(mpd) > 0:
                        raise __pe(
                            f"Already gave {KEY_PRODUCTION_TIME}"
                            f"{KEY_IDX_START}{station_id}{CSV_SEPARATOR}"
                            f"{product_id}{KEY_IDX_END}", oline, line_idx)
                    mpd.extend(
                        map(float, str.split(value, CSV_SEPARATOR)))
            else:
                infos[key] = value

    if name is None:
        raise ValueError(f"Did not specify instance name ({KEY_NAME}).")
    if n_products is None:
        raise ValueError("Did not specify instance n_products"
                         f" ({KEY_N_PRODUCTS}).")
    if n_customers is None:
        raise ValueError("Did not specify instance n_customers"
                         f" ({KEY_N_CUSTOMERS}).")
    if n_stations is None:
        raise ValueError("Did not specify instance n_stations"
                         f" ({KEY_N_STATIONS}).")
    if n_demands is None:
        raise ValueError("Did not specify instance n_demands"
                         f" ({KEY_N_DEMANDS}).")
    if time_end_warmup is None:
        raise ValueError("Did not specify instance time_end_warmup"
                         f" ({KEY_TIME_END_WARMUP}).")
    if time_end_measure is None:
        raise ValueError("Did not specify instance time_end_measure"
                         f" ({KEY_TIME_END_MEASURE}).")
    if routes is None:
        raise ValueError(f"Did not specify instance routes ({KEY_ROUTE}).")
    if demands is None:
        raise ValueError(f"Did not specify instance demands ({KEY_DEMAND}).")
    if in_warehouse is None:
        raise ValueError("Did not specify instance warehouse values"
                         f" ({KEY_IN_WAREHOUSE}).")
    if station_product_times is None:
        raise ValueError("Did not specify per-station product production"
                         f"times ({KEY_PRODUCTION_TIME}).")

    return Instance(name, n_products, n_customers, n_stations, n_demands,
                    time_end_warmup, time_end_measure,
                    routes, demands, in_warehouse, station_product_times,
                    infos)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def compute_finish_time(start_time: float, amount: int,
                        production_times: np.ndarray) -> float:
    """
    Compute the time when one job is finished.

    The production times are cyclic intervals of unit production times and
    interval ends.

    :param start_time: the starting time of the job
    :param amount: the number of units to be produced
    :param production_times: the production times array
    :return: the end time

    Here, the production time is 10 time units / 1 product unit, valid until
    end time 100.
    >>> compute_finish_time(0.0, 1, np.array((10.0, 100.0), np.float64))
    10.0

    Here, the production time is 10 time units / 1 product unit, valid until
    end time 100. We begin producing at time unit 250. Since the production
    periods are cyclic, this is OK: we would be halfway through the third
    production period when the request comes in. It will consume 10 time units
    and be done at time unit 260.
    >>> compute_finish_time(250.0, 1, np.array((10.0, 100.0)))
    260.0

    Here, the end time of the production time validity is at time unit 100.
    However, we begin producing 1 product unit at time step 90. This unit will
    use 10 time units, meaning that its production is exactly finished when
    the production time validity ends.
    It will be finished at time step 100.
    >>> compute_finish_time(90.0, 1, np.array((10.0, 100.0)))
    100.0

    Here, the end time of the production time validity is at time unit 100.
    However, we begin producing 1 product unit at time step 95. This unit would
    use 10 time units. It will use these units, even though this extends beyond
    the end of the production time window.
    >>> compute_finish_time(95.0, 1, np.array((10.0, 100.0)))
    105.0

    Now we have two production periods. The production begins again at time
    step 95. It will use 10 time units, even though this extends into the
    second period.
    >>> compute_finish_time(95.0, 1, np.array((10.0, 100.0, 20.0, 200.0)))
    105.0

    Now things get more complex. We want to do 10 units of product.
    We start in the first period, so one unit will be completed there.
    This takes the starting time for the next job to 105, which is in the
    second period. Here, one unit of product takes 20 time units. We can
    finish producing one unit until time 125 and start the production of a
    second one, taking until 145. Now the remaining three units are produced
    until time 495
    >>> compute_finish_time(95.0, 10, np.array((
    ...     10.0, 100.0, 20.0, 140.0, 50.0, 5000.0)))
    495.0
    >>> 95 + (1*10 + 2*20 + 7*50)
    495

    We again produce 10 product units starting at time step 95. The first one
    takes 10 time units, taking us into the second production interval at time
    105. Then we can again do two units here, which consume 40 time units,
    taking us over the edge into the third interval at time unit 145. Here we
    do two units using 50 time units. We ahen are at time 245, which wraps back
    to 45. So the remaining 5 units take 10 time units each.
    >>> compute_finish_time(95.0, 10, np.array((
    ...     10.0, 100.0, 20.0, 140.0, 50.0, 200.0)))
    295.0
    >>> 95 + (1*10 + 2*20 + 2*50 + 5*10)
    295

    This is the same as the last example, but this time, the last interval
    (3 time units until 207) is skipped over by the long production of the
    second 50-time-unit product.
    >>> compute_finish_time(95.0, 10, np.array((
    ...     10.0, 100.0, 20.0, 140.0, 50.0, 200.0,  3.0, 207.0)))
    295.0
    >>> 95 + (1*10 + 2*20 + 2*50 + 5*10)
    295

    Production unit times may extend beyond the intervals.
    >>> compute_finish_time(0.0, 5, np.array((1000.0, 100.0, 10.0, 110.0)))
    5000.0
    >>>
    5 * 1000
    """
    time_mod: Final[float] = production_times[-1]
    low_end: Final[int] = len(production_times)
    total: Final[int] = low_end // 2

    # First, we need to find the segment in the production cycle
    # where the production begins. We use a binary search for that.
    remaining: int | float = amount
    seg_start: float = start_time % time_mod
    low: int = 0
    high: int = total
    while low < high:
        mid: int = (low + high) // 2
        th: float = production_times[mid * 2 + 1]
        if th <= seg_start:
            low = mid + 1
        else:
            high = mid - 1
    low *= 2

    # Now we can cycle through the production cycle until the product has
    # been produced.
    while True:
        max_time = production_times[low + 1]
        while max_time <= seg_start:
            low += 2
            if low >= low_end:
                low = 0
                seg_start = 0.0
            max_time = production_times[low + 1]

        unit_time = production_times[low]
        can_do: int = ceil(min(
            max_time - seg_start, remaining) / unit_time)
        duration = can_do * unit_time
        seg_start += duration
        start_time += duration
        remaining -= can_do
        if remaining <= 0:
            return float(start_time)



"""
Some utilities for random sampling.

The goal that we follow with class :class:`Distribution` is to have
clearly defined integer-producing random distributions. We want to be
able to say exactly how to generate some random numbers.

A distribution can be sampled using the method :meth:`~Distribution.sample`.
Each distribution has a mean value, which either may be an exact value or
an approximate result, that can be obtained via :meth:`~Distribution.mean`.
Sometimes, distributions can be simplified, which is supported by
:meth:`~Distribution.simplify`.

>>> from moptipy.utils.nputils import rand_generator
>>> from statistics import mean
>>> rnd = rand_generator(0)

>>> const = Const(12.3)
>>> const
Const(v=12.3)
>>> const.mean()
12.3
>>> const.sample(rnd)
12.3

>>> normal = Normal(1, 2.0)
>>> normal
Normal(mu=1, sd=2)
>>> x = [normal.sample(rnd) for _ in range(200)]
>>> x[:20]
[1.2514604421867865, 0.7357902734173962, 2.280845300886564,\
 1.2098002343060794, -0.07133874632222192, 1.7231901098189695,\
 3.6080000902602745, 2.8941619262584846, -0.4074704716139852,\
 -1.530842942092105, -0.24654892507470438, 1.0826519586944872,\
 -3.6500615492776687, 0.5624166721349085, -1.4918218945061303,\
 -0.4645347094069032, -0.08851796571461978, 0.3673996872616909,\
 1.8232610727482657, 3.085026738885355]
>>> mean(x) / normal.mean()
1.0305262793198813

>>> exponential = Exponential(3)
>>> x = [exponential.sample(rnd) for _ in range(200)]
>>> x[:20]
[1.3203895772033505, 1.152983246827425, 4.527545171626064,\
 5.080441711712409, 0.6498200242245252, 3.408652958826374,\
 2.842677620245357, 2.1132194336022, 3.5838508479713393,\
 1.2223825336486978, 2.8454397976498504, 0.8653980789905962,\
 0.19572166348792452, 3.2506725793229854, 0.9426446058446336,\
 3.246902386754473, 10.294412603282666, 5.275683923067543,\
 0.49517363091492944, 0.2982336402218482]
>>> mean(x) / exponential.mean()
1.088857332669081

>>> gamma = Gamma(3.0, 0.26)
>>> gamma
Gamma(k=3, theta=0.26)
>>> x = [gamma.sample(rnd) for _ in range(200)]
>>> x[:20]
[1.1617763743292708, 1.305755109480284, 0.7627948403389954,\
 1.2735522637897285, 0.7742951665621697, 1.074233520618276,\
 0.6324661100546898, 1.4627037699922791, 0.5739033567160827,\
 0.5555065636904546, 0.629236234283296, 0.3666171387296996,\
 0.3780976936750937, 0.9511433672028858, 1.2607313263258062,\
 1.4442096466925938, 0.48758642085808085, 1.247724803721524,\
 1.9359140456080306, 1.3935246884396764]
>>> mean(x) / gamma.mean()
0.962254253258444

>>> Gamma.from_alpha_beta(2, 0.5)
Erlang(k=2, theta=2)

>>> Gamma.from_alpha_beta(2.5, 0.5)
Gamma(k=2.5, theta=2)

>>> Erlang.from_alpha_beta(2, 0.5)
Erlang(k=2, theta=2)

>>> erlang = Erlang(1.0, 0.26)
>>> erlang
Erlang(k=1, theta=0.26)
>>> x = [erlang.sample(rnd) for _ in range(200)]
>>> x[:20]
[0.29911329228981776, 0.3630768060267626, 0.14111385731543394,\
 0.0745673280234536, 0.029950507989979877, 0.04741877104350835,\
 0.38599089026561223, 0.047919114170390194, 0.06921557868837301,\
 0.4066084140331242, 0.07170887998378667, 0.022061870233843223,\
 0.04904717644388396, 0.32082097064821674, 0.001884448999141546,\
 0.6687964040577958, 0.060598863807579915, 0.21491377996577304,\
 0.23088301776258766, 0.23667780086315618]
>>> mean(x) / erlang.mean()
0.9212193955320179

>>> uniform = Uniform(10.5, 17)
>>> uniform
Uniform(low=10.5, high=17)
>>> x = [uniform.sample(rnd) for _ in range(200)]
>>> x[:20]
[11.235929257903324, 13.89459856391694, 14.196833332199233,\
 13.871456027849515, 14.485310406702393, 16.20468251006836,\
 13.7773314921396, 12.964459911549145, 12.167722566772781,\
 12.49450317595701, 14.145245846833722, 15.669920217581366,\
 13.367286694558258, 10.764955127505994, 11.723004274328007,\
 11.089239517262232, 12.666732191193558, 14.948448277461127,\
 14.339645757381653, 14.803829334728565]
>>> mean(x) / uniform.mean()
0.9968578545459643

>>> choice = Choice((Const(2), gamma, normal))
>>> choice
Choice(ch=(Const(v=2), Gamma(k=3, theta=0.26), Normal(mu=1, sd=2)))
>>> x = [choice.sample(rnd) for _ in range(200)]
>>> x[:20]
[0.5499364190576, -0.10428920251005325, 2, 2, 0.44263544084840273,\
 0.6088189450771303, 2, 2, 2, 2, -0.32003290715104904, 2,\
 0.6165299227577784, 1.0445083345086352, 2, 2, 2, -0.5970322857539738,\
 -1.6672705710198277, 2]
>>> mean(x) / choice.mean()
0.9849730453309948

>>> lower = AtLeast(2, normal)
>>> lower
AtLeast(lb=2, d=Normal(mu=1, sd=2))
>>> x = [lower.sample(rnd) for _ in range(200)]
>>> x[:20]
[2.9715155848474772, 3.407503493034132, 2.59634461519561,\
 2.0472458472897714, 3.9865670827840334, 2.015117730344058,\
 2.0316584714999935, 5.36625737470408, 3.042848158343226,\
 3.390407444638451, 4.5503184215686, 3.7682459882073007,\
 2.0539760253651305, 2.134886147372958, 2.500182239395479,\
 2.891402111997337, 2.7393826524907228, 2.3449577842364766,\
 2.9043074694017195, 4.7173482582723825]
>>> mean(x) / lower.mean()
1.0178179732075965

>>> interval = In(1, 10, gamma)
>>> interval
In(lb=1, ub=10, d=Gamma(k=3, theta=0.26))
>>> x = [interval.sample(rnd) for _ in range(200)]
>>> x[:20]
[1.3488423972739083, 1.0929631399361601, 1.681162621901135,\
 1.655614926246918, 1.041948891842002, 2.0773395958990175,\
 1.3338891374921853, 1.2478964188175743, 1.9417070894505217,\
 1.3990572178987266, 1.1216118870312337, 1.0641160239253207,\
 2.131253600219639, 1.0337453883221577, 1.396499618416345,\
 1.9865175145136038, 1.2555473269031396, 1.4412027583435465,\
 1.320740351247919, 1.0407411942999665]
>>> mean(x) / interval.mean()
1.0256329849020747

>>> erlang2 = Gamma.from_k_and_mean(3, 10)
>>> erlang2
Erlang(k=3, theta=3.3333333333333335)
>>> erlang2.mean()
10
>>> x = [erlang2.sample(rnd) for _ in range(200)]
>>> x[:20]
[10.087506399523226, 12.928131914870168, 12.330250639007767,\
 5.305123692562998, 21.085037136404374, 6.6603691824173135,\
 2.961302890492059, 9.810557147180853, 8.051620919921454,\
 8.750329405836668, 3.9511445189935763, 5.570300668751883,\
 16.70132947692463, 7.831425379483914, 11.154757962484842,\
 8.78943102381046, 8.395847820234795, 16.42251602814587,\
 17.1628992966332, 9.684008648356015]
>>> mean(x) / erlang2.mean()
1.022444624140316
"""

from dataclasses import dataclass
from math import fsum, isfinite
from typing import Callable, Final, cast

from moptipy.utils.nputils import rand_generator, rand_seeds_from_str
from numpy.random import Generator
from pycommons.math.int_math import try_int
from pycommons.types import type_error

#: the maximum number of trials during a sampling process
_MAX_TRIALS: int = 1_000_000


class Distribution:
    """A base class for distributions."""

    def sample(self, random: Generator) -> int | float:
        """
        Sample a random number following this distribution generator.

        Each call to this function returns exactly one number.

        :param random: the random number generator
        :return: the number
        """
        raise NotImplementedError

    def simplify(self) -> "Distribution":
        """
        Try to simplify this distribution.

        Some distributions can trivially be simplified. For example, if you
        have applied a range limit (:class:`In`) to a constant distribution
        (class:`Const`), then this can be simplified to just the constant.
        If such simplification is possible, this method returns the simplified
        distribution. Otherwise, it just returns the distribution itself.

        :returns: a simplified version of this distribution
        """
        if not isinstance(self, Distribution):
            raise type_error(self, "self", Distribution)
        return self

    def mean(self) -> int | float:
        """
        Get the mean or approximate mean of the distribution.

        Some distribution overwrite this method to produce an exact computed
        expected value or mean. This default implementation just computes the
        arithmetic mean of 10'000 samples of the distribution. This serves as
        baseline approximation for any case where a closed form mathematical
        definition of the expected value is not available.

        :return: the mean or approximated mean
        """
        sample: Callable[[Generator], int | float] = self.sample
        gen: Final[Generator] = rand_generator(rand_seeds_from_str(
            repr(self), 1)[0])
        return try_int(fsum(sample(gen) for _ in range(10_000)) / 10_000)


@dataclass(order=True, frozen=True)
class Const(Distribution):
    """A constant value."""

    #: the constant value
    v: int | float

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        object.__setattr__(self, "v", try_int(self.v))

    def sample(self, random: Generator) -> int | float:
        """
        Sample the constant integer.

        :param random: the random number generator
        :return: the integer
        """
        return self.v

    def mean(self) -> int | float:
        """
        Get the mean of this distribution.

        :return: the mean
        """
        return self.v


@dataclass(order=True, frozen=True)
class Normal(Distribution):
    """A class representing a normal distribution."""

    #: the expected value and center of the distribution
    mu: int | float
    #: the standard deviation of the distribution
    sd: int | float

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        if not (isfinite(self.mu) and isfinite(self.sd) and (self.sd > 0)):
            raise ValueError(f"Invalid parameters {self}.")
        object.__setattr__(self, "mu", try_int(self.mu))
        object.__setattr__(self, "sd", try_int(self.sd))

    def sample(self, random: Generator) -> float:
        """
        Sample from the normal distribution.

        :param random: the random number generator
        :return: the result
        """
        return random.normal(self.mu, self.sd)

    def mean(self) -> int | float:
        """
        Get the mean of this distribution.

        :return: the mean
        """
        return self.mu


@dataclass(order=True, frozen=True)
class Exponential(Distribution):
    """A class representing an exponential distribution."""

    #: the exponential distribution parameter
    eta: int | float

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        object.__setattr__(self, "eta", try_int(self.eta))
        if self.eta <= 0:
            raise ValueError(f"Invalid setup {self!r}.")

    def sample(self, random: Generator) -> float:
        """
        Sample from the Exponential distribution.

        :param random: the random number generator
        :return: the result
        """
        return random.exponential(self.eta)

    def mean(self) -> int | float:
        """
        Get the mean of this distribution.

        :return: the mean
        """
        return try_int(self.eta)


@dataclass(order=True, frozen=True)
class Gamma(Distribution):
    """
    A class representing a Gamma distribution.

    Here, `k` is the shape and `theta` is the scale parameter.
    If you use a parameterization with `alpha` and `beta`, you need to create
    the distribution using :meth:`~Gamma.from_alpha_beta` instead. The reason
    is that `shape = 1/beta`, see
    https://www.statlect.com/probability-distributions/gamma-distribution.
    """

    #: the shape parameter
    k: int | float
    #: the scale parameter
    theta: int | float

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        object.__setattr__(self, "k", try_int(self.k))
        object.__setattr__(self, "theta", try_int(self.theta))
        if not (self.k > 0) and (self.theta > 0):
            raise ValueError(f"Invalid parameters {self}.")

    def sample(self, random: Generator) -> float:
        """
        Sample from the Gamma distribution.

        :param random: the random number generator
        :return: the result
        """
        return random.gamma(self.k, self.theta)

    def simplify(self) -> "Distribution":
        """
        Try to simplify this distribution.

        A Gamma distribution may simplify to either an :class:`Erlang` or an
        :class:`Exponential` distribution, depending on its parameters.
        If the :attr:`~Gamma.k` is `1`, then it is actually an
        :class:`Exponential` distribution. If :attr:`~Gamma.k` is an integer,
        then the distribution is an :class:`Erlang` distribution.

        1. https://www.statisticshowto.com/gamma-distribution
        2. https://www.statisticshowto.com/erlang-distribution

        :returns: a simplified version of this distribution
        """
        return Exponential(self.k) if self.k == 1 else (
            Erlang(self.k, self.theta) if isinstance(
                self.k, int) and not isinstance(self, Erlang) else self)

    def mean(self) -> int | float:
        """
        Get the mean of this distribution.

        :return: the mean
        """
        return try_int(self.k * self.theta)

    @classmethod
    def from_alpha_beta(cls, alpha: int | float, beta: int | float) \
            -> "Distribution":
        """
        Create a Gamma distribution from `alpha` and `beta`.

        :param alpha: the alpha parameter
        :param beta: the beta parameter
        :return: the distribution
        """
        beta = try_int(beta)
        if beta == 0:
            raise ValueError(f"beta cannot be {beta}.")
        return cls(alpha, 1 / beta).simplify()

    @classmethod
    def from_k_and_mean(cls, k: int | float, mean: int | float) \
            -> "Distribution":
        """
        Create the Gamma distribution from the value of `k` and a mean.

        :param k: the shape parameter
        :param mean: the mean
        :return: the distribution
        """
        k = try_int(k)
        mean = try_int(mean)
        if (mean <= 0) or (k <= 0):
            raise ValueError(f"Invalid values k={k}, mean={mean}.")
        return Gamma(k, mean / k).simplify()


class Erlang(Gamma):
    """The Erlang distribution."""

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        super().__post_init__()
        if not isinstance(self.k, int):
            raise type_error(self.k, "k", int)


@dataclass(order=True, frozen=True)
class Uniform(Distribution):
    """A class representing a uniform distribution."""

    #: the lowest permitted value
    low: int | float
    #: the highest permitted value
    high: int | float

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        object.__setattr__(self, "low", try_int(self.low))
        object.__setattr__(self, "high", try_int(self.high))
        if self.high <= self.low:
            raise ValueError(f"Invalid parameters {self}.")

    def sample(self, random: Generator) -> float:
        """
        Sample from the uniform distribution.

        :param random: the random number generator
        :return: the result
        """
        return random.uniform(self.low, self.high)

    def mean(self) -> int | float:
        """
        Get the mean of this distribution.

        :return: the mean
        """
        return try_int((self.high + self.low) / 2)


@dataclass(order=True, frozen=True)
class Choice(Distribution):
    """
    A class representing a uniform choice.

    >>> Choice((Uniform(1, 2), Uniform(3, 4))).simplify()
    Choice(ch=(Uniform(low=1, high=2), Uniform(low=3, high=4)))

    >>> Choice((Uniform(1, 2), Uniform(1.0, 2))).simplify()
    Uniform(low=1, high=2)

    >>> Choice((Uniform(1, 2), Choice(
    ...     (Const(1), Uniform(1, 2))))).simplify()
    Choice(ch=(Uniform(low=1, high=2), Const(v=1), Uniform(low=1, high=2)))
    """

    #: the choices
    ch: tuple[Distribution, ...]

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        if not isinstance(self.ch, tuple):
            raise TypeError(f"Invalid types {self}.")
        for v in self.ch:
            if not isinstance(v, Distribution):
                raise type_error(v, "choice", Distribution)

    def mean(self) -> int | float:
        """
        Get the mean of this distribution.

        :return: the mean
        """
        return try_int(fsum(d.mean() for d in self.ch) / tuple.__len__(
            self.ch))

    def sample(self, random: Generator) -> int | float:
        """
        Sample from the uniform distribution.

        :param random: the random number generator
        :return: the result
        """
        return self.ch[random.integers(tuple.__len__(self.ch))].sample(random)

    def simplify(self) -> Distribution:
        """
        Try to simplify this distribution.

        :returns: a simplified version of this distribution
        """
        ch: Final[tuple[Distribution, ...]] = self.ch
        if tuple.__len__(ch) <= 1:
            return ch[0].simplify()
        done: list[Distribution] = []
        needs: bool = False
        for dist in ch:
            use: Distribution = dist.simplify()
            if use != dist:
                needs = True
            if isinstance(use, Choice):
                needs = True
                done.extend(use.ch)
            else:
                done.append(use)

        total: int = list.__len__(done)
        dc: Distribution = done[0]
        if total <= 1:
            return dc

        all_same: bool = True
        for oth in done:
            if oth != dc:
                all_same = False
                break
        return dc if all_same else (Choice(tuple(done)) if needs else self)


@dataclass(order=True, frozen=True)
class AtLeast(Distribution):
    """
    A distribution that is lower-bounded.

    >>> AtLeast(5, Const(7))
    AtLeast(lb=5, d=Const(v=7))

    >>> AtLeast(5, AtLeast(8, Const(17)))
    AtLeast(lb=8, d=Const(v=17))

    >>> AtLeast(8, AtLeast(5, Const(17)))
    AtLeast(lb=8, d=Const(v=17))
    """

    #: the inclusive lower bound
    lb: int | float
    #: the inner distribution to sample from
    d: Distribution

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        object.__setattr__(self, "lb", try_int(self.lb))
        dd: Final[Distribution] = self.d
        if isinstance(dd, Const):
            if cast("Const", dd).v < self.lb:
                raise ValueError(f"Invalid distribution {self!r}.")
        elif isinstance(dd, AtLeast):
            dlb: AtLeast = cast("AtLeast", dd)
            object.__setattr__(self, "lb", max(self.lb, dlb.lb))
            object.__setattr__(self, "d", dlb.d)
        if isinstance(dd, In):
            idd: In = cast("In", dd)
            ulb: int | float = max(idd.lb, self.lb)
            if ulb >= idd.ub:
                raise ValueError(f"Invalid distribution {self!r}.")

    def simplify(self) -> "Distribution":
        """
        Try to simplify this distribution.

        :returns: a simplified version of this distribution
        """
        dd: Final[Distribution] = self.d
        if isinstance(dd, Const):
            return dd
        if isinstance(dd, In):
            idd: In = cast("In", dd)
            return In(max(idd.lb, self.lb), idd.ub, idd.d).simplify()
        if (self.lb <= 0) and (isinstance(dd, Exponential | Gamma | Erlang)):
            return dd
        return self

    def sample(self, random: Generator) -> int | float:
        """
        Sample from the lower-bounded distribution.

        :param random: the random number generator
        :return: the result
        """
        s: Final[Callable[[Generator], int | float]] = self.d.sample
        lb: Final[int | float] = self.lb
        for _ in range(_MAX_TRIALS):
            v = s(random)
            if lb <= v:
                return v
        raise ValueError(f"Failed to sample from {self!r}.")


@dataclass(order=True, frozen=True)
class In(Distribution):
    """
    A distribution that is lower and upper-bounded.

    >>> In(1, 10, Const(6))
    In(lb=1, ub=10, d=Const(v=6))

    >>> In(1, 10, In(5, 12, Const(6)))
    In(lb=5, ub=10, d=Const(v=6))

    >>> In(1, 10, AtLeast(6, Const(6)))
    In(lb=6, ub=10, d=Const(v=6))
    """

    #: the inclusive lower bound
    lb: int | float
    #: the exclusive upper bound
    ub: int | float
    #: the inner distribution to sample from
    d: Distribution

    def __post_init__(self) -> None:
        """Perform some basic sanity checks and cleanup."""
        object.__setattr__(self, "lb", try_int(self.lb))
        object.__setattr__(self, "ub", try_int(self.ub))
        if not isinstance(self.d, Distribution):
            raise TypeError(f"Invalid types {self}.")
        if self.ub <= self.lb:
            raise ValueError(f"Invalid range {self}.")
        dd: Distribution = self.d
        lb: Final[int | float] = self.lb
        ub: Final[int | float] = self.ub
        if isinstance(dd, In):
            idd: In = cast("In", dd)
            ulb: int | float = max(idd.lb, lb)
            uub: int | float = min(idd.ub, ub)
            if ulb >= uub:
                raise ValueError(f"Invalid distribution {self!r}.")
            object.__setattr__(self, "lb", ulb)
            object.__setattr__(self, "ub", uub)
            dd = idd.d
            object.__setattr__(self, "d", dd)
        elif isinstance(dd, AtLeast):
            ldd: AtLeast = cast("AtLeast", dd)
            ulb = max(ldd.lb, lb)
            if ulb >= ub:
                raise ValueError(f"Invalid distribution {self!r}.")
            object.__setattr__(self, "lb", ulb)
            dd = ldd.d
            object.__setattr__(self, "d", dd)
        elif isinstance(dd, Uniform):  # fix a uniform distribution
            udd: Uniform = cast("Uniform", dd)
            ulb = max(udd.low, lb)
            uub = min(udd.high, ub)
            if ulb >= uub:
                raise ValueError(f"Invalid distribution {self!r}.")
            if (ulb != udd.low) or (uub != udd.high):
                object.__setattr__(self, "d", Uniform(ulb, uub))
            object.__setattr__(self, "lb", ulb)
            object.__setattr__(self, "ub", uub)
        if isinstance(dd, Const) and not lb <= cast("Const", dd).v < ub:
            raise ValueError(f"Invalid distribution {self!r}.")

    def sample(self, random: Generator) -> int | float:
        """
        Sample from the lower-bounded distribution.

        :param random: the random number generator
        :return: the result
        """
        s: Final[Callable[[Generator], int | float]] = self.d.sample
        lb: Final[int | float] = self.lb
        ub: Final[int | float] = self.ub
        for _ in range(_MAX_TRIALS):
            v = s(random)
            if lb <= v < ub:
                return v
        raise ValueError(f"Failed to sample from {self!r}.")

    def simplify(self) -> Distribution:
        """
        Simplify this distribution.

        :return: the simplified distribution

        >>> In(-3, 4, Const(3)).simplify()
        Const(v=3)
        >>> ii = In(-10, 10, Uniform(3, 20))
        >>> ii
        In(lb=3, ub=10, d=Uniform(low=3, high=10))
        >>> ii.simplify()
        Uniform(low=3, high=10)
        """
        return self.d if isinstance(self.d, Const | Uniform) else self

    def mean(self) -> int | float:
        """
        Get the mean of this distribution.

        :return: the mean
        """
        return self.d.mean() if isinstance(self.d, Const | Uniform) \
            else super().mean()


def distribution(d: int | float | Distribution) -> Distribution:
    """
    Get the distribution from the parameter.

    :param d: the integer value or distribution
    :return: the canonicalized distribution

    >>> distribution(7)
    Const(v=7)

    >>> distribution(3.4)
    Const(v=3.4)

    >>> distribution(Choice((Const(4.0), )))
    Const(v=4)
    """
    if isinstance(d, int | float):
        return Const(d)
    if isinstance(d, Distribution):
        old_d: Distribution | None = None
        while old_d is not d:
            old_d = d
            d = d.simplify()
            if not isinstance(d, Distribution):
                break
        return d
    raise type_error(d, "d", (Distribution, int))



"""
Methods for generating instances.

Here we provide some basic utilities for generating deterministic variants of
instances such as those used in the paper [1] by Thürer et al.

>>> from moptipyapps.utils.sampling import Gamma
>>> inst = sample_mfc_instance([
...  Product(0, (0, 1), Gamma.from_alpha_beta(3, 0.26))], [
...  Station(0, Gamma.from_k_and_mean(3, 10)),
...  Station(1, Gamma.from_k_and_mean(2, 10))],
...  time_end_measure=100, seed=123)

>>> inst.name
'mfc_1_2_100_0x7b'

>>> inst.n_demands
7

>>> inst.demands
(Demand(arrival=5.213885878801001, deadline=5.213885878801001, demand_id=0,\
 customer_id=0, product_id=0, amount=1, measure=False),\
 Demand(arrival=25.872387132411287, deadline=25.872387132411287, demand_id=1,\
 customer_id=1, product_id=0, amount=1, measure=False),\
 Demand(arrival=43.062182155666896, deadline=43.062182155666896, demand_id=2,\
 customer_id=2, product_id=0, amount=1, measure=True),\
 Demand(arrival=49.817978344678004, deadline=49.817978344678004, demand_id=3,\
 customer_id=3, product_id=0, amount=1, measure=True),\
 Demand(arrival=58.21166922638016, deadline=58.21166922638016, demand_id=4,\
 customer_id=4, product_id=0, amount=1, measure=True),\
 Demand(arrival=69.09054693162531, deadline=69.09054693162531, demand_id=5,\
 customer_id=5, product_id=0, amount=1, measure=True),\
 Demand(arrival=88.804579148131, deadline=88.804579148131, demand_id=6,\
 customer_id=6, product_id=0, amount=1, measure=True))

>>> len(inst.station_product_unit_times[0][0])
1600

>>> len(inst.station_product_unit_times[1][0])
1600

>>> inst.time_end_measure
100.0

>>> inst.time_end_warmup
30.0

>>> d = dict(inst.infos)
>>> del d["info_generated_on"]
>>> del d["info_generator_version"]
>>> d
{'info_generator': 'moptipyapps.prodsched.mfc_generator',\
 'info_rand_seed_src': 'USER_PROVIDED',\
 'info_rand_seed': '0x7b',\
 'info_time_end_measure_src': 'USER_PROVIDED',\
 'info_time_end_measure': '100',\
 'info_time_end_warmup_src': 'SAMPLED',\
 'info_time_end_warmup': '30',\
 'info_name_src': 'SAMPLED',\
 'info_product_interarrival_times[0]':\
 'Erlang(k=3, theta=3.846153846153846)',\
 'info_product_route[0]': 'USER_PROVIDED',\
 'info_station_processing_time[0]':\
 'Erlang(k=3, theta=3.3333333333333335)',\
 'info_station_processing_time_window_length[0]': 'Const(v=0.125)',\
 'info_station_processing_time[1]': 'Erlang(k=2, theta=5)',\
 'info_station_processing_time_window_length[1]': 'Const(v=0.125)'}

>>> inst = sample_mfc_instance(seed=23445)
>>> inst.name
'mfc_10_13_10000_0x5b95'

>>> inst.n_demands
9922

>>> len([dem for dem in inst.demands if dem.product_id == 0])
959

>>> len([dem for dem in inst.demands if dem.product_id == 1])
1055

>>> [len(k[0]) for k in inst.station_product_unit_times]
[160000, 160000, 0, 160000, 0, 0, 0, 0, 160000, 160000, 160000, 0, 0]

1. Matthias Thürer, Nuno O. Fernandes, Hermann Lödding, and Mark Stevenson.
   Material Flow Control in Make-to-Stock Production Systems: An Assessment of
   Order Generation, Order Release and Production Authorization by Simulation
   Flexible Services and Manufacturing Journal. 37(1):1-37. March 2025.
   doi:https://doi.org/10.1007/s10696-024-09532-2
"""
import datetime
from dataclasses import dataclass
from typing import Callable, Final, Iterable

from moptipy.utils.nputils import (
    rand_generator,
    rand_seed_generate,
)
from moptipy.utils.strings import sanitize_name
from numpy.random import Generator
from pycommons.math.int_math import try_int
from pycommons.strings.string_conv import num_to_str
from pycommons.types import check_int_range, type_error

from moptipyapps.prodsched.instance import (
    KEY_IDX_END,
    KEY_IDX_START,
    MAX_VALUE,
    Demand,
    Instance,
)
from moptipyapps.utils.sampling import (
    Const,
    Distribution,
    Erlang,
    Gamma,
    Uniform,
    distribution,
)
from moptipyapps.version import __version__

#: the "now" function
__DTN: Final[Callable[[], datetime.datetime]] = datetime.datetime.now

#: the maximum trials for sampling
MAX_TRIALS: Final[int] = 1_000_000

#: The information key for interarrival times
INFO_PRODUCT_INTERARRIVAL_TIME_DIST: Final[str] = \
    "info_product_interarrival_times"

#: The generator key
INFO_GENERATOR: Final[str] = "info_generator"
#: The generator version
INFO_GENERATOR_VERSION: Final[str] = "info_generator_version"
#: When was the instance generated?
INFO_GENERATED_ON: Final[str] = "info_generated_on"
#: The information key for interarrival times
INFO_PRODUCT_ROUTE: Final[str] = "info_product_route"
#: a fixed structure
INFO_USER_PROVIDED: Final[str] = "USER_PROVIDED"
#: a sampled structure
INFO_SAMPLED: Final[str] = "SAMPLED"
#: The information key for the processing time distribution
INFO_STATION_PROCESSING_TIME: Final[str] = "info_station_processing_time"
#: The information key for the processing time window length distribution
INFO_STATION_PROCESSING_WINDOW_LENGTH: Final[str] = \
    "info_station_processing_time_window_length"
#: the random seed
INFO_RAND_SEED: Final[str] = "info_rand_seed"
#: the random seed source
INFO_RAND_SEED_SRC: Final[str] = f"{INFO_RAND_SEED}_src"
#: the name source
INFO_NAME_SRC: Final[str] = "info_name_src"
#: the warmup time end
INFO_TIME_END_WARMUP: Final[str] = "info_time_end_warmup"
#: the source of the warmup time end
INFO_TIME_END_WARMUP_SRC: Final[str] = f"{INFO_TIME_END_WARMUP}_src"
#: the measurement time end
INFO_TIME_END_MEASURE: Final[str] = "info_time_end_measure"
#: the source of the measurement time end
INFO_TIME_END_MEASURE_SRC: Final[str] = f"{INFO_TIME_END_MEASURE}_src"


@dataclass(order=True, frozen=True)
class Product:
    """The product sampling definition."""

    #: the product ID
    product_id: int
    #: the routing of the product
    routing: tuple[int, ...]
    #: the interarrival distribution
    interarrival_times: Distribution

    def __init__(self, product_id: int, routing: Iterable[int],
                 interarrival_times: int | float | Distribution) -> None:
        """
        Create the product sampling instruction.

        :param product_id: the product id
        :param routing: the routing information
        :param interarrival_times: the interarrival time distribution
        """
        object.__setattr__(self, "product_id", check_int_range(
            product_id, "product_id", 0, 1_000_000))
        route: tuple[int, ...] = tuple(routing)
        n_route: int = tuple.__len__(route)
        if n_route <= 0:
            raise ValueError("Route cannot be empty!")
        for k in route:
            check_int_range(k, "station", 0, 1_000_000)
        object.__setattr__(self, "routing", route)
        object.__setattr__(
            self, "interarrival_times", distribution(interarrival_times))

    def log_info(self, infos: dict[str, str]) -> None:
        """
        Log the sampling information of this product to the infos `dict`.

        :param infos: the information dictionary
        """
        key: Final[str] = f"{KEY_IDX_START}{self.product_id}{KEY_IDX_END}"
        infos[f"{INFO_PRODUCT_INTERARRIVAL_TIME_DIST}{key}"] = repr(
            self.interarrival_times)
        infos[f"{INFO_PRODUCT_ROUTE}{key}"] = INFO_USER_PROVIDED


@dataclass(order=True, frozen=True)
class Station:
    """The station sampling definition."""

    #: the product ID
    station_id: int
    #: the processing time distribution
    processing_time: Distribution
    #: the processing window distribution
    processing_windows: Distribution

    def __init__(
            self, station_id: int,
            processing_time: int | float | Distribution,
            processing_windows: int | float | Distribution | None = None) \
            -> None:
        """
        Create the station sampling instruction.

        :param station_id: the station id
        :param processing_time: the processing time distribution
        :param processing_windows: the processing time window length
            distribution
        """
        object.__setattr__(self, "station_id", check_int_range(
            station_id, "station_id", 0, 1_000_000))
        object.__setattr__(
            self, "processing_time", distribution(processing_time))

        if processing_windows is None:
            processing_windows = Const(1 / 8)
        object.__setattr__(
            self, "processing_windows", distribution(
                processing_windows))

    def log_info(self, infos: dict[str, str]) -> None:
        """
        Log the sampling information of this product to the infos `dict`.

        :param infos: the information dictionary
        """
        key: Final[str] = f"{KEY_IDX_START}{self.station_id}{KEY_IDX_END}"
        infos[f"{INFO_STATION_PROCESSING_TIME}{key}"] = repr(
            self.processing_time)
        infos[f"{INFO_STATION_PROCESSING_WINDOW_LENGTH}{key}"] = repr(
            self.processing_windows)


# pylint: disable=R0914,R0912,R0915
def sample_mfc_instance(products: Iterable[Product] | None = None,
                        stations: Iterable[Station] | None = None,
                        time_end_warmup: int | float | None = None,
                        time_end_measure: int | float | None = None,
                        name: str | None = None,
                        seed: int | None = None) -> Instance:
    """
    Sample an MFC instance.

    :param products: the products
    :param stations: the work stations
    :param time_end_warmup: the end of the warmup period
    :param time_end_measure: the end of the measurement period
    :param name: the instance name
    :param seed: the random seed, if any
    :return: the instance
    """
    generator: str = str(__file__)
    idx: int = str.rfind(generator, "moptipyapps")
    if idx >= 0:
        generator = str.removesuffix(generator[idx:].replace("/", "."), ".py")
    else:
        generator = "mfc_generator"
    infos: Final[dict[str, str]] = {
        INFO_GENERATOR: generator,
        INFO_GENERATOR_VERSION: __version__,
        INFO_GENERATED_ON: str(__DTN()),
    }

    if products is None:
        products = default_products()
    products = sorted(products)
    n_products: Final[int] = list.__len__(products)
    if n_products <= 0:
        raise ValueError(f"Cannot have {n_products} products.")

    ids: Final[set[int]] = set()
    used_stations: Final[set[int]] = set()
    for product in products:
        if not isinstance(product, Product):
            raise type_error(product, "product", Product)
        ids.add(product.product_id)
        used_stations.update(product.routing)
    if (set.__len__(ids) != n_products) or (
            max(ids) - min(ids) + 1 != n_products):
        raise ValueError("Inconsistent product ids.")

    n_stations: Final[int] = set.__len__(used_stations)
    if not 0 < n_stations < 1_000_000:
        raise ValueError(f"Invalid number {n_stations} of stations.")
    if stations is None:
        if n_stations == 13:
            stations = default_stations()
        else:
            raise ValueError(
                "Can only use default settings with 13 stations, "
                f"but got {n_stations}.")

    stations = sorted(stations)
    n_stations_real: int = list.__len__(stations)
    if n_stations_real != n_stations:
        raise ValueError(
            f"Products use {n_stations} stations,"
            f" but {n_stations_real} are provided.")

    ids.clear()
    for station in stations:
        if not isinstance(station, Station):
            raise type_error(station, "station", Station)
        ids.add(station.station_id)
    min_id: int = min(ids)
    max_id: int = max(ids)
    if (set.__len__(ids) != n_stations) or (
            max_id - min_id + 1 != n_stations):
        raise ValueError("Inconsistent station ids.")
    if ids != used_stations:
        raise ValueError(
            f"Station ids are {min_id}...{max_id}, but products use "
            f"stations {sorted(used_stations)}.")

    if seed is None:
        infos[INFO_RAND_SEED_SRC] = INFO_SAMPLED
        seed = rand_seed_generate()
    else:
        infos[INFO_RAND_SEED_SRC] = INFO_USER_PROVIDED
    if not isinstance(seed, int):
        raise type_error(seed, "seed", int)
    infos[INFO_RAND_SEED] = hex(seed)

    if time_end_measure is None:
        time_end_measure = 10_000 if (time_end_warmup is None) or (
            time_end_warmup <= 0) else max(
            time_end_warmup + 1, (10 * time_end_warmup) / 3)
        infos[INFO_TIME_END_MEASURE_SRC] = INFO_SAMPLED
    else:
        infos[INFO_TIME_END_MEASURE_SRC] = INFO_USER_PROVIDED
    time_end_measure = try_int(time_end_measure)
    if not 0 < time_end_measure < MAX_VALUE:
        raise ValueError(
            f"Invalid time_end_measure={time_end_measure}.")
    infos[INFO_TIME_END_MEASURE] = num_to_str(time_end_measure)

    if time_end_warmup is None:
        time_end_warmup = (3 * time_end_measure) / 10
        infos[INFO_TIME_END_WARMUP_SRC] = INFO_SAMPLED
    else:
        infos[INFO_TIME_END_WARMUP_SRC] = INFO_USER_PROVIDED
    time_end_warmup = try_int(time_end_warmup)
    if not 0 <= time_end_warmup < time_end_measure:
        raise ValueError(f"Invalid time_end_warmup={time_end_warmup} "
                         f"for time_end_measure={time_end_measure}.")
    infos[INFO_TIME_END_WARMUP] = num_to_str(time_end_warmup)

    if name is None:
        infos[INFO_NAME_SRC] = INFO_SAMPLED
        s: str = num_to_str(time_end_measure).replace(".", "d")
        name = f"mfc_{n_products}_{n_stations}_{s}_{seed:#x}"
    else:
        infos[INFO_NAME_SRC] = INFO_USER_PROVIDED
    uname: str = sanitize_name(name)
    if uname != name:
        raise ValueError(f"Invalid name {name!r}.")

    random: Final[Generator] = rand_generator(seed)

    # sample the demands
    demands: Final[list[Demand]] = []
    current_id: int = 0
    trials: int = MAX_TRIALS
    for product in products:
        time: float = 0.0
        while True:
            while True:
                until = product.interarrival_times.sample(random)
                if until > 0:
                    break
                trials -= 1
                if trials <= 0:
                    raise ValueError(
                        "Failed to sample appropriate inter-arrival time"
                        f" from {product.interarrival_times}")
            time += until
            if time >= time_end_measure:
                break
            demands.append(Demand(
                arrival=time, deadline=time, demand_id=current_id,
                customer_id=current_id, product_id=product.product_id,
                amount=1, measure=time_end_warmup <= time))
            current_id += 1

    #: sample the working times
    production_times: Final[list[list[list[float]]]] = []
    for station in stations:
        times: list[float] = []
        time = 0.0
        while True:
            while True:
                processing = station.processing_time.sample(random)
                if processing > 0:
                    break
                trials -= 1
                if trials <= 0:
                    raise ValueError(
                        "Failed to sample appropriate processing time"
                        f" from {station.processing_time}")
            while True:
                window = station.processing_windows.sample(random)
                if window > 0:
                    break
                trials -= 1
                if trials <= 0:
                    raise ValueError(
                        "Failed to sample appropriate processing window time"
                        f" from {station.processing_windows}")
            time += window
            times.extend((processing, time))
            if time >= time_end_measure:
                break
        production_times.append([
            times if station.station_id in product.routing else []
            for product in products])

    # log the information
    for product in products:
        product.log_info(infos)
    for station in stations:
        station.log_info(infos)

    return Instance(
        name=name, n_products=n_products,
        n_customers=current_id, n_stations=n_stations,
        n_demands=current_id,
        time_end_warmup=time_end_warmup, time_end_measure=time_end_measure,
        routes=(product.routing for product in products),
        demands=demands, warehous_at_t0=[0] * n_products,
        station_product_unit_times=production_times,
        infos=infos)


def __s1t0(s: Iterable[int]) -> tuple[int, ...]:
    """
    Convert stations from 1 to 0-based index.

    :param s: the stations
    :return: the index

    >>> __s1t0((1, 2, 3))
    (0, 1, 2)
    """
    return tuple(x - 1 for x in s)


def default_products() -> tuple[Product, ...]:
    """
    Create the default product sequence as used in [1].

    :return: the default product sequence

    >>> default_products()
    (Product(product_id=0, routing=(0, 1, 3, 1, 8, 9, 10), \
interarrival_times=Erlang(k=3, theta=3.3333333333333335)), \
Product(product_id=1, routing=(0, 1, 4, 1, 7, 8, 9, 10), \
interarrival_times=Erlang(k=2, theta=5)), \
Product(product_id=2, routing=(0, 1, 5, 3, 1, 8, 11, 10), \
interarrival_times=Uniform(low=5, high=15)), \
Product(product_id=3, routing=(0, 1, 6, 3, 1, 8, 9, 10), \
interarrival_times=Erlang(k=3, theta=3.3333333333333335)), \
Product(product_id=4, routing=(0, 1, 3, 11, 1, 8, 1, 12), \
interarrival_times=Erlang(k=4, theta=2.5)), \
Product(product_id=5, routing=(0, 1, 4, 11, 1, 8, 6, 12), \
interarrival_times=Erlang(k=2, theta=5)), \
Product(product_id=6, routing=(0, 1, 5, 11, 1, 7, 1, 12), \
interarrival_times=Erlang(k=4, theta=2.5)), \
Product(product_id=7, routing=(0, 1, 2, 6, 3, 11, 1, 7, 5, 8, 1, 12), \
interarrival_times=Uniform(low=5, high=15)), \
Product(product_id=8, routing=(0, 1, 2, 4, 3, 5, 11, 1, 7, 1, 9, 5, 12), \
interarrival_times=Erlang(k=4, theta=2.5)), \
Product(product_id=9, routing=(0, 1, 2, 5, 1, 3, 11, 6, 1, 8, 10, 4, 12), \
interarrival_times=Erlang(k=2, theta=5)))

    1. Matthias Thürer, Nuno O. Fernandes, Hermann Lödding, and Mark
       Stevenson. Material Flow Control in Make-to-Stock Production Systems:
       An Assessment of Order Generation, Order Release and Production
       Authorization by Simulation Flexible Services and Manufacturing
       Journal. 37(1):1-37. March 2025.
       doi:https://doi.org/10.1007/s10696-024-09532-2
    """
    return (
        Product(0, __s1t0((1, 2, 4, 2, 9, 10, 11)),
                Erlang.from_k_and_mean(3, 10)),
        Product(1, __s1t0((1, 2, 5, 2, 8, 9, 10, 11)),
                Erlang.from_k_and_mean(2, 10)),
        Product(2, __s1t0((1, 2, 6, 4, 2, 9, 12, 11)),
                Uniform(5, 15)),
        Product(3, __s1t0((1, 2, 7, 4, 2, 9, 10, 11)),
                Erlang.from_k_and_mean(3, 10)),
        Product(4, __s1t0((1, 2, 4, 12, 2, 9, 2, 13)),
                Erlang.from_k_and_mean(4, 10)),
        Product(5, __s1t0((1, 2, 5, 12, 2, 9, 7, 13)),
                Erlang.from_k_and_mean(2, 10)),
        Product(6, __s1t0((1, 2, 6, 12, 2, 8, 2, 13)),
                Erlang.from_k_and_mean(4, 10)),
        Product(7, __s1t0((1, 2, 3, 7, 4, 12, 2, 8, 6, 9, 2, 13)),
                Uniform(5, 15)),
        Product(8, __s1t0((1, 2, 3, 5, 4, 6, 12, 2, 8, 2, 10, 6, 13)),
                Erlang.from_k_and_mean(4, 10)),
        Product(9, __s1t0((1, 2, 3, 6, 2, 4, 12, 7, 2, 9, 11, 5, 13)),
                Erlang.from_k_and_mean(2, 10)))


def default_stations() -> tuple[Station, ...]:
    """
    Create the default station sequence as used in [1].

    :return: the default product station

    >>> default_stations()
    (Station(station_id=0, processing_time=Erlang(k=3, theta=0.26),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=1, processing_time=Erlang(k=3, theta=0.12),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=2, processing_time=Erlang(k=2, theta=1.33),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=3, processing_time=Exponential(eta=1),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=4, processing_time=Erlang(k=3, theta=0.67),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=5, processing_time=Erlang(k=4, theta=0.35),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=6, processing_time=Erlang(k=3, theta=0.59),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=7, processing_time=Erlang(k=3, theta=0.63),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=8, processing_time=Erlang(k=2, theta=0.59),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=9, processing_time=Erlang(k=3, theta=0.6),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=10, processing_time=Exponential(eta=1),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=11, processing_time=Erlang(k=4, theta=0.29),\
 processing_windows=Const(v=0.125)),\
 Station(station_id=12, processing_time=Erlang(k=3, theta=0.48),\
 processing_windows=Const(v=0.125)))

    1. Matthias Thürer, Nuno O. Fernandes, Hermann Lödding, and Mark
       Stevenson. Material Flow Control in Make-to-Stock Production Systems:
       An Assessment of Order Generation, Order Release and Production
       Authorization by Simulation Flexible Services and Manufacturing
       Journal. 37(1):1-37. March 2025.
       doi:https://doi.org/10.1007/s10696-024-09532-2
    """
    return (
        Station(0, Gamma(3, 0.26)),
        Station(1, Gamma(3, 0.12)),
        Station(2, Gamma(2, 1.33)),
        Station(3, Gamma(1, 1.06)),
        Station(4, Gamma(3, 0.67)),
        Station(5, Gamma(4, 0.35)),
        Station(6, Gamma(3, 0.59)),
        Station(7, Gamma(3, 0.63)),
        Station(8, Gamma(2, 0.59)),
        Station(9, Gamma(3, 0.6)),
        Station(10, Gamma(1, 1.44)),
        Station(11, Gamma(4, 0.29)),
        Station(12, Gamma(3, 0.48)))
